import * as i0 from '@angular/core';
import { InjectionToken, inject, ElementRef, NgZone, Injectable, Inject, EventEmitter, Component, ChangeDetectionStrategy, HostBinding, Input, Output, ViewChild, NgModule, Pipe, Directive, Optional, ChangeDetectorRef, Self, SkipSelf, ContentChild } from '@angular/core';
import * as THREE from 'three';
import * as i5 from '@angular/common';
import { DOCUMENT, CommonModule, NgForOf } from '@angular/common';
import { __rest } from 'tslib';
import * as i1 from 'rxjs';
import { Observable, Subject, pipe, debounceTime, takeUntil, fromEvent, share, ReplaySubject, take, filter, map, distinctUntilChanged, combineLatest, isObservable, of, tap, concatMap, scheduled, queueScheduler, withLatestFrom, startWith, noop, Subscription, asapScheduler, merge, mapTo, BehaviorSubject, defer, forkJoin, catchError, throwError } from 'rxjs';

const NGT_PARENT_OBJECT = new InjectionToken('THREE_OBJECT_3D as Parent');
function createParentObjectProvider(existing, factory) {
    return {
        provide: NGT_PARENT_OBJECT,
        useFactory: (existing) => {
            return () => factory === null || factory === void 0 ? void 0 : factory(existing);
        },
        deps: [existing],
    };
}
const NGT_HOST_PARENT_OBJECT = new InjectionToken('The Host THREE_OBJECT_3D as Parent. Used for abstract objects like Billboard, Text...');
function createHostParentObjectProvider(existing) {
    return {
        provide: NGT_HOST_PARENT_OBJECT,
        useFactory: (existing) => {
            return () => { var _a; return (_a = existing.parentObjectFn) === null || _a === void 0 ? void 0 : _a.call(existing); };
        },
        deps: [existing],
    };
}

const NGT_OBJECT_TYPE = new InjectionToken('Object Type', {
    providedIn: 'root',
    factory: () => THREE.Object3D,
});
const NGT_OBJECT_POST_INIT = new InjectionToken('Object PostInit', {
    providedIn: 'root',
    factory: () => undefined,
});

const NGT_PERFORMANCE_OPTIONS = new InjectionToken('NgtPerformance options', {
    providedIn: 'root',
    factory: () => ({
        current: 1,
        min: 0.5,
        max: 1,
        debounce: 200,
    }),
});

const NGT_RESIZE_BOX_DEFAULT = 'content-box';
const NGT_RESIZE_DEBOUNCE_DEFAULT = 200;
const NGT_RESIZE_OPTIONS = new InjectionToken('ngtResizeObserver Options', {
    providedIn: 'root',
    factory: () => ({
        box: NGT_RESIZE_BOX_DEFAULT,
        scroll: false,
        offsetSize: false,
        debounce: NGT_RESIZE_DEBOUNCE_DEFAULT,
    }),
});
const NGT_RESIZE_OBSERVER_SUPPORT = new InjectionToken('Resize Observer API support', {
    providedIn: 'root',
    factory: () => {
        const document = inject(DOCUMENT);
        return (!!document.defaultView &&
            'ResizeObserver' in document.defaultView &&
            document.defaultView['ResizeObserver'] != null);
    },
});

const DEFAULT_NGT_CANVAS_OPTIONS = {
    production: true,
    initialLog: false,
    projectContent: false,
};
const NGT_CANVAS_OPTIONS = new InjectionToken('Canvas Options', {
    providedIn: 'root',
    factory: () => DEFAULT_NGT_CANVAS_OPTIONS,
});
function provideCanvasOptions(options) {
    return {
        provide: NGT_CANVAS_OPTIONS,
        useValue: Object.assign(Object.assign({}, DEFAULT_NGT_CANVAS_OPTIONS), (options || {})),
    };
}

const NGT_IS_WEBGL_AVAILABLE = new InjectionToken('isWebGLAvailable', {
    providedIn: 'root',
    factory: () => {
        var _a;
        const document = inject(DOCUMENT);
        try {
            const canvas = document.createElement('canvas');
            return !!(((_a = document.defaultView) === null || _a === void 0 ? void 0 : _a.WebGL2RenderingContext) &&
                canvas.getContext('webgl2'));
        }
        catch (e) {
            return false;
        }
    },
});

class NgtResize extends Observable {
    constructor({ nativeElement }, zone, document, isSupport, { box, offsetSize, scroll, debounce }) {
        let observer;
        let lastBounds;
        let lastEntries = [];
        const torndown$ = new Subject();
        const scrollContainers = findScrollContainers(nativeElement);
        // set actual debounce values early, so effects know if they should react accordingly
        const scrollDebounce = debounce
            ? typeof debounce === 'number'
                ? debounce
                : debounce.scroll
            : null;
        const resizeDebounce = debounce
            ? typeof debounce === 'number'
                ? debounce
                : debounce.resize
            : null;
        const debounceAndDestroy = (debounce) => {
            return pipe(debounceTime(debounce !== null && debounce !== void 0 ? debounce : 0), takeUntil(torndown$));
        };
        super((subscriber) => {
            if (!isSupport) {
                subscriber.error('ResizeObserver is not supported in your browser. Please use a polyfill');
                return;
            }
            zone.runOutsideAngular(() => {
                const callback = (entries) => {
                    lastEntries = entries;
                    const { left, top, width, height, bottom, right, x, y } = nativeElement.getBoundingClientRect();
                    const size = {
                        left,
                        top,
                        width,
                        height,
                        bottom,
                        right,
                        x,
                        y,
                    };
                    if (nativeElement instanceof HTMLElement && offsetSize) {
                        size.height = nativeElement.offsetHeight;
                        size.width = nativeElement.offsetWidth;
                    }
                    Object.freeze(size);
                    subscriber.next(Object.assign({ entries, dpr: document.defaultView.devicePixelRatio }, size));
                    if (!areBoundsEqual(lastBounds || {}, size)) {
                        lastBounds = size;
                    }
                };
                const boundEntriesCallback = () => {
                    callback(lastEntries);
                };
                observer = new ResizeObserver(callback);
                observer.observe(nativeElement, { box });
                if (scroll) {
                    if (scrollContainers) {
                        scrollContainers.forEach((scrollContainer) => {
                            fromEvent(scrollContainer, 'scroll', {
                                capture: true,
                                passive: true,
                            })
                                .pipe(debounceAndDestroy(scrollDebounce))
                                .subscribe(boundEntriesCallback);
                        });
                    }
                    fromEvent(document.defaultView, 'scroll', {
                        capture: true,
                        passive: true,
                    })
                        .pipe(debounceAndDestroy(scrollDebounce))
                        .subscribe(boundEntriesCallback);
                }
                fromEvent(document.defaultView, 'resize')
                    .pipe(debounceAndDestroy(resizeDebounce))
                    .subscribe(boundEntriesCallback);
            });
            return () => {
                if (observer) {
                    observer.unobserve(nativeElement);
                    observer.disconnect();
                }
                torndown$.next();
                torndown$.complete();
            };
        });
        return this.pipe(debounceTime(scrollDebounce || 0), share({
            connector: () => new ReplaySubject(1),
            resetOnRefCountZero: true,
            resetOnComplete: true,
        }));
    }
}
NgtResize.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtResize, deps: [{ token: ElementRef }, { token: NgZone }, { token: DOCUMENT }, { token: NGT_RESIZE_OBSERVER_SUPPORT }, { token: NGT_RESIZE_OPTIONS }], target: i0.ɵɵFactoryTarget.Injectable });
NgtResize.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtResize });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtResize, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: i0.ElementRef, decorators: [{
                        type: Inject,
                        args: [ElementRef]
                    }] }, { type: i0.NgZone, decorators: [{
                        type: Inject,
                        args: [NgZone]
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_RESIZE_OBSERVER_SUPPORT]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_RESIZE_OPTIONS]
                    }] }];
    } });
// Returns a list of scroll offsets
function findScrollContainers(element) {
    const result = [];
    if (!element || element === document.body)
        return result;
    const { overflow, overflowX, overflowY } = window.getComputedStyle(element);
    if ([overflow, overflowX, overflowY].some((prop) => prop === 'auto' || prop === 'scroll'))
        result.push(element);
    return [
        ...result,
        ...findScrollContainers(element.parentElement),
    ];
}
// Checks if element boundaries are equal
const keys = [
    'x',
    'y',
    'top',
    'bottom',
    'left',
    'right',
    'width',
    'height',
];
const areBoundsEqual = (a, b) => keys.every((key) => a[key] === b[key]);

function checkNeedsUpdate(value) {
    if (value !== null &&
        typeof value === 'object' &&
        'needsUpdate' in value) {
        value['needsUpdate'] = true;
    }
}
function applyProps(instance, props) {
    if (!Object.keys(props).length)
        return;
    if ('__ngt' in props) {
        instance['__ngt'] = props['__ngt'];
    }
    const unknownInstance = instance;
    if (unknownInstance['set'] != null &&
        typeof unknownInstance['set'] === 'function' &&
        !(unknownInstance instanceof THREE.Raycaster)) {
        try {
            unknownInstance['set'](props);
        }
        catch (e) {
            console.info(`Swallowing erroneous "set" invoked on ${unknownInstance.constructor.name} as non fatal: ${e}`);
        }
    }
    for (const [key, prop] of Object.entries(props)) {
        // raycast is null or undefined. we'll skip
        if (key === 'raycast' && prop == undefined)
            continue;
        const currentInstance = unknownInstance;
        const targetProp = currentInstance[key];
        if (targetProp &&
            targetProp['set'] &&
            (targetProp['copy'] || targetProp instanceof THREE.Layers)) {
            if (Array.isArray(prop)) {
                if (targetProp['fromArray'])
                    targetProp['fromArray'](prop);
                else
                    targetProp['set'](...prop);
            }
            else if (targetProp['copy'] &&
                prop &&
                prop.constructor &&
                targetProp.constructor.name ===
                    prop.constructor.name) {
                targetProp['copy'](prop);
            }
            else if (prop !== undefined) {
                const isColor = targetProp instanceof THREE.Color;
                // Allow setting array scalars
                if (!isColor && targetProp['setScalar'])
                    targetProp['setScalar'](prop);
                // Layers have no copy function, we must therefore copy the mask property
                else if (targetProp instanceof THREE.Layers &&
                    prop instanceof THREE.Layers)
                    targetProp.mask = prop.mask;
                // Otherwise just set ...
                else
                    targetProp['set'](prop);
                // Auto-convert sRGB colors, for now ...
                // https://github.com/pmndrs/react-three-fiber/issues/344
                if (!currentInstance['linear'] && isColor)
                    targetProp.convertSRGBToLinear();
            }
        }
        else {
            currentInstance[key] = prop;
            // Auto-convert sRGB textures, for now ...
            // https://github.com/pmndrs/react-three-fiber/issues/344
            if (!currentInstance['linear'] &&
                currentInstance[key] instanceof THREE.Texture)
                currentInstance[key].encoding =
                    THREE.sRGBEncoding;
        }
        checkNeedsUpdate(prop);
        checkNeedsUpdate(targetProp);
    }
}

function calculateDpr(dpr) {
    return Array.isArray(dpr)
        ? Math.min(Math.max(dpr[0], window.devicePixelRatio), dpr[1])
        : dpr;
}

function isOrthographicCamera(def) {
    return def && def.isOrthographicCamera;
}

class NgtStore {
    constructor() {
        // Should be used only in ngOnDestroy.
        this.destroySubject$ = new ReplaySubject(1);
        // Exposed to any extending Store to be used for the teardown.
        this.destroy$ = this.destroySubject$.asObservable();
        this.stateSubject$ = new ReplaySubject(1);
    }
    /** Completes all relevant Observable streams. */
    ngOnDestroy() {
        this.stateSubject$.complete();
        this.destroySubject$.next();
    }
    get(projector) {
        let value;
        this.stateSubject$
            .pipe(take(1), skipUndefined(), map((state) => (projector ? projector(state) : state)), skipUndefined())
            .subscribe((result) => {
            value = result;
        });
        return value;
    }
    /**
     * Patches the state with provided partial state.
     *
     * @param partialStateOrUpdaterFn a partial state or a partial updater
     * function that accepts the state and returns the partial state.
     */
    set(partialStateOrUpdaterFn) {
        const patchedState = typeof partialStateOrUpdaterFn === 'function'
            ? partialStateOrUpdaterFn(this.get())
            : partialStateOrUpdaterFn;
        this.update((state, partialState) => (Object.assign(Object.assign({}, state), partialState)))(patchedState);
    }
    select(...args) {
        if (args.length === 0) {
            return this.stateSubject$.pipe(skipUndefined(), distinctUntilChanged(), share({
                connector: () => new ReplaySubject(1),
                resetOnComplete: true,
                resetOnRefCountZero: true,
                resetOnError: true,
            }), takeUntil(this.destroy$));
        }
        const { observables, projector } = processSelectorArgs(args);
        let observable$;
        // If there are no Observables to combine, then we'll just map the value.
        if (observables.length === 0) {
            observable$ = this.stateSubject$.pipe(skipUndefined(), map(projector));
        }
        else {
            // If there are multiple arguments, then we're aggregating selectors, so we need
            // to take the combineLatest of them before calling the map function.
            observable$ = combineLatest(observables).pipe(map((projectorArgs) => projector(...projectorArgs)));
        }
        return observable$.pipe(skipUndefined(), distinctUntilChanged(), share({
            connector: () => new ReplaySubject(1),
            resetOnComplete: true,
            resetOnRefCountZero: true,
            resetOnError: true,
        }), takeUntil(this.destroy$));
    }
    /**
     * Creates an effect.
     *
     * This effect is subscribed to throughout the lifecycle of the ComponentStore.
     * @param generator A function that takes an origin Observable input and
     *     returns an Observable. The Observable that is returned will be
     *     subscribed to for the life of the component.
     * @return A function that, when called, will trigger the origin Observable.
     */
    effect(generator) {
        const origin$ = new Subject();
        generator(origin$)
            // tied to the lifecycle 👇 of ComponentStore
            .pipe(takeUntil(this.destroy$))
            .subscribe();
        return ((observableOrValue) => {
            const observable$ = isObservable(observableOrValue)
                ? observableOrValue
                : of(observableOrValue);
            return observable$
                .pipe(takeUntil(this.destroy$))
                .subscribe((value) => {
                // any new 👇 value is pushed into a stream
                origin$.next(value);
            });
        });
    }
    onCanvasReady(ready$, callback, useEffect = false) {
        const operator = useEffect ? tapEffect : tap;
        return this.effect(operator(() => callback()))(ready$);
    }
    update(updaterFn) {
        return ((observableOrValue) => {
            const observable$ = isObservable(observableOrValue)
                ? observableOrValue
                : of(observableOrValue);
            return observable$
                .pipe(concatMap((value) => scheduled([value], queueScheduler).pipe(withLatestFrom(this.stateSubject$.pipe(startWith({}))))), takeUntil(this.destroy$))
                .subscribe({
                next: ([value, currentState]) => {
                    this.stateSubject$.next(updaterFn((currentState || {}), value));
                },
                error: (error) => {
                    this.stateSubject$.error(error);
                },
            });
        });
    }
}
NgtStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgtStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtStore, decorators: [{
            type: Injectable
        }] });
/**
 * An extended `tap` operator that accepts an `effectFn` which:
 * - runs on every `next` notification from `source$`
 * - can optionally return a `cleanUp` function that
 * invokes from the 2nd `next` notification onward and on `unsubscribe` (destroyed)
 *
 *
 * @example
 * ```typescript
 * source$.pipe(
 *  tapEffect((sourceValue) = {
 *    const cb = () => {
 *      doStuff(sourceValue);
 *    };
 *    addListener('event', cb);
 *
 *    return () => {
 *      removeListener('event', cb);
 *    }
 *  })
 * )
 * ```
 */
function tapEffect(effectFn) {
    let cleanupFn = noop;
    let firstRun = false;
    let prev = undefined;
    const teardown = (error) => {
        return () => {
            if (cleanupFn) {
                cleanupFn({ prev, complete: true, error });
            }
        };
    };
    return tap({
        next: (value) => {
            if (cleanupFn && firstRun) {
                cleanupFn({ prev, complete: false, error: false });
            }
            const cleanUpOrVoid = effectFn(value);
            if (cleanUpOrVoid) {
                cleanupFn = cleanUpOrVoid;
            }
            prev = value;
            if (!firstRun) {
                firstRun = true;
            }
        },
        complete: teardown(false),
        unsubscribe: teardown(false),
        error: teardown(true),
    });
}
function debounceSync() {
    return (source) => new Observable((observer) => {
        let actionSubscription;
        let actionValue;
        const rootSubscription = new Subscription();
        rootSubscription.add(source.subscribe({
            complete: () => {
                if (actionSubscription) {
                    observer.next(actionValue);
                }
                observer.complete();
            },
            error: (error) => {
                observer.error(error);
            },
            next: (value) => {
                actionValue = value;
                if (!actionSubscription) {
                    actionSubscription = asapScheduler.schedule(() => {
                        observer.next(actionValue);
                        actionSubscription = undefined;
                    });
                    rootSubscription.add(actionSubscription);
                }
            },
        }));
        return rootSubscription;
    });
}
function skipUndefined() {
    return filter((value) => value !== undefined);
}
function startWithUndefined() {
    return startWith(undefined);
}
function processSelectorArgs(args) {
    const selectorArgs = Array.from(args);
    // Last argument is either projector or config
    const projector = selectorArgs.pop();
    // The Observables to combine, if there are any.
    const observables = selectorArgs;
    return {
        observables,
        projector,
    };
}

const position = new THREE.Vector3();
const defaultTarget = new THREE.Vector3();
class NgtCanvasStore extends NgtStore {
    constructor(performance, { nativeElement: { clientHeight, clientWidth }, }, resizeResult$, document) {
        var _a, _b, _c;
        super();
        this.resizeResult$ = resizeResult$;
        this.ready$ = this.select((s) => s.ready).pipe(filter((ready) => ready));
        this.camera$ = this.select((s) => s.camera);
        this.scene$ = this.select((s) => s.scene);
        this.renderer$ = this.select((s) => s.renderer);
        this.raycaster$ = this.select((s) => s.raycaster);
        this.allConstructed$ = this.select(this.camera$, this.scene$, this.renderer$, this.raycaster$, (camera, scene, renderer, raycaster) => !!camera && !!renderer && !!scene && !!raycaster).pipe(map((ready) => ({ ready })));
        this.dimensions$ = this.select(this.select((s) => s.size), this.select((s) => s.viewport), (size, viewport) => ({ size, viewport }));
        this.init = this.effect(tap((canvasElement) => {
            this.set(this.allConstructed$);
            this.updateDimensions(this.dimensions$);
            this.resize(this.resizeResult$);
            this.initRenderer(canvasElement);
        }));
        this.resize = this.effect(tap(({ width, height, dpr }) => {
            this.set({
                size: { width, height },
                viewport: Object.assign(Object.assign({}, this.get((s) => s.viewport)), { dpr }),
            });
        }));
        this.updateDimensions = this.effect(tap(({ size, viewport }) => {
            const { camera, renderer, ready, cameraOptions } = this.get();
            if (ready) {
                // update renderer
                renderer.setPixelRatio(viewport.dpr);
                renderer.setSize(size.width, size.height);
                // leave the userland camera alone
                if (cameraOptions instanceof THREE.Camera)
                    return;
                if (isOrthographicCamera(camera)) {
                    camera.left = size.width / -2;
                    camera.right = size.width / 2;
                    camera.top = size.height / 2;
                    camera.bottom = size.height / -2;
                }
                else {
                    camera.aspect = size.width / size.height;
                }
                camera.updateProjectionMatrix();
                // https://github.com/pmndrs/react-three-fiber/issues/178
                // Update matrix world since the renderer is a frame late
                camera.updateMatrixWorld();
            }
        }));
        this.initRenderer = this.effect(tapEffect((canvasElement) => {
            const { size, viewport, vr, linear, flat, orthographic, shadows, glOptions, sceneOptions, cameraOptions, raycasterOptions, } = this.get();
            // Scene
            const scene = new THREE.Scene();
            applyProps(scene, sceneOptions);
            // Camera
            const isCamera = cameraOptions instanceof THREE.Camera;
            const camera = isCamera
                ? cameraOptions
                : orthographic
                    ? new THREE.OrthographicCamera(0, 0, 0, 0, 0.1, 1000)
                    : new THREE.PerspectiveCamera(75, size.width / size.height, 0.1, 1000);
            if (!isCamera) {
                camera.position.z = 5;
                if (cameraOptions) {
                    applyProps(camera, cameraOptions);
                    // Update projection matrix after applying props
                    camera.updateProjectionMatrix();
                }
                // look at center if initial rotation isn't set
                if (!(cameraOptions === null || cameraOptions === void 0 ? void 0 : cameraOptions.rotation))
                    camera.lookAt(0, 0, 0);
            }
            // Raycaster
            const raycaster = new THREE.Raycaster();
            const _a = raycasterOptions || {}, { params } = _a, options = __rest(_a, ["params"]);
            applyProps(raycaster, Object.assign(Object.assign({ enabled: true }, options), { params: Object.assign(Object.assign({}, raycaster.params), (params || {})) }));
            // Renderer
            const customRenderer = (typeof glOptions === 'function'
                ? glOptions(canvasElement)
                : glOptions);
            // userland custom renderer, assign as-is
            if ((customRenderer === null || customRenderer === void 0 ? void 0 : customRenderer.render) != null) {
                this.set({
                    renderer: customRenderer,
                    scene,
                    camera,
                    raycaster: raycaster,
                });
            }
            else {
                const renderer = new THREE.WebGLRenderer(Object.assign({ powerPreference: 'high-performance', canvas: canvasElement, antialias: true, alpha: true }, (glOptions || {})));
                if (glOptions) {
                    applyProps(renderer, glOptions);
                }
                if (shadows) {
                    renderer.shadowMap.enabled = true;
                    if (typeof shadows === 'object') {
                        Object.assign(renderer.shadowMap, shadows);
                    }
                    else {
                        renderer.shadowMap.type = THREE.PCFSoftShadowMap;
                    }
                }
                if (!linear) {
                    // auto color management
                    renderer.toneMapping = THREE.ACESFilmicToneMapping;
                    renderer.outputEncoding = THREE.sRGBEncoding;
                }
                else {
                    renderer.outputEncoding = THREE.LinearEncoding;
                }
                if (flat) {
                    renderer.toneMapping = THREE.NoToneMapping;
                }
                renderer.setClearAlpha(0);
                renderer.setPixelRatio(calculateDpr(viewport.dpr));
                renderer.setSize(size.width, size.height);
                if (vr) {
                    renderer.xr.enabled = true;
                }
                this.set({
                    renderer,
                    scene,
                    camera,
                    raycaster: raycaster,
                });
            }
            return () => {
                const { renderer, vr } = this.get();
                if (renderer) {
                    renderer.renderLists.dispose();
                    renderer.forceContextLoss();
                    if (vr) {
                        renderer.setAnimationLoop(null);
                    }
                }
            };
        }));
        this.set({
            ready: false,
            vr: false,
            orthographic: false,
            flat: false,
            linear: false,
            size: {
                width: clientWidth,
                height: clientHeight,
            },
            mouse: new THREE.Vector2(),
            clock: new THREE.Clock(),
            frameloop: 'always',
            performance,
            objects: {},
            dpr: ((_a = document.defaultView) === null || _a === void 0 ? void 0 : _a.devicePixelRatio) || 1,
            shadows: false,
            cameraOptions: {},
            glOptions: {},
            raycasterOptions: {},
            sceneOptions: {},
            viewport: {
                initialDpr: ((_b = document.defaultView) === null || _b === void 0 ? void 0 : _b.devicePixelRatio) || 1,
                dpr: ((_c = document.defaultView) === null || _c === void 0 ? void 0 : _c.devicePixelRatio) || 1,
                width: clientWidth,
                height: clientHeight,
                aspect: clientWidth / clientHeight,
                distance: 0,
                factor: 0,
                getCurrentViewport: (camera, target = defaultTarget, size) => {
                    const { camera: defaultCamera, size: defaultSize } = this.get();
                    if (!camera) {
                        camera = defaultCamera;
                    }
                    if (!size) {
                        size = defaultSize;
                    }
                    const { width, height } = size;
                    const aspect = width / height;
                    const distance = camera
                        .getWorldDirection(position)
                        .distanceTo(target);
                    if (isOrthographicCamera(camera)) {
                        return {
                            width: width / camera.zoom,
                            height: height / camera.zoom,
                            factor: 1,
                            distance,
                            aspect,
                        };
                    }
                    const fov = (camera.fov * Math.PI) / 180; // convert vertical fov to radians
                    const h = 2 * Math.tan(fov / 2) * distance; // height of viewport
                    const w = h * aspect; // width of viewport
                    return {
                        width: w,
                        height: h,
                        factor: width / w,
                        distance,
                        aspect,
                    };
                },
            },
            controls: null,
        });
    }
    get isLinear() {
        return this.get((s) => s.linear);
    }
}
NgtCanvasStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCanvasStore, deps: [{ token: NGT_PERFORMANCE_OPTIONS }, { token: i0.ElementRef }, { token: NgtResize }, { token: DOCUMENT }], target: i0.ɵɵFactoryTarget.Injectable });
NgtCanvasStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCanvasStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCanvasStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () {
        return [{ type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_PERFORMANCE_OPTIONS]
                    }] }, { type: i0.ElementRef }, { type: i1.Observable, decorators: [{
                        type: Inject,
                        args: [NgtResize]
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }];
    } });

function makeVector2(input) {
    if (!input)
        return undefined;
    if (input instanceof THREE.Vector2) {
        return input;
    }
    if (Array.isArray(input)) {
        return new THREE.Vector2(...input);
    }
    return new THREE.Vector2(input, input);
}
function makeVector3(input) {
    if (!input)
        return undefined;
    if (input instanceof THREE.Vector3) {
        return input;
    }
    if (Array.isArray(input)) {
        return new THREE.Vector3(...input);
    }
    return new THREE.Vector3(input, input, input);
}
function makeVector4(input) {
    if (!input)
        return undefined;
    if (input instanceof THREE.Vector4) {
        return input;
    }
    if (Array.isArray(input)) {
        return new THREE.Vector4(...input);
    }
    return new THREE.Vector4(input, input, input, input);
}
function makeForSet(type, input) {
    if (!input)
        return undefined;
    if (input instanceof type) {
        return input;
    }
    return new type(...input);
}
function makeColor(color) {
    if (!color)
        return undefined;
    if (color instanceof THREE.Color) {
        return color;
    }
    if (Array.isArray(color)) {
        return new THREE.Color(...color);
    }
    return new THREE.Color(color);
}
const idCache = {};
/**
 * Generate a random Id or an Event's uuid
 *
 * @param {NgtIntersection} event
 *
 * @internal
 * @private
 */
function makeId(event) {
    if (event) {
        return ((event.eventObject || event.object).uuid +
            '/' +
            event.index +
            event.instanceId);
    }
    const newId = THREE.MathUtils.generateUUID();
    // ensure not already used
    if (!idCache[newId]) {
        idCache[newId] = true;
        return newId;
    }
    return makeId();
}

class NgtAnimationFrameStore extends NgtStore {
    constructor() {
        super();
        this.registerAnimation = this.effect(tapEffect((_a) => {
            var { uuid } = _a, record = __rest(_a, ["uuid"]);
            if (uuid) {
                this.set((state) => ({
                    animations: new Map(state.animations).set(uuid, record),
                }));
            }
            return ({ prev: { uuid: prevUuid } = {}, complete }) => {
                if (prevUuid !== uuid || complete) {
                    this.unregister(uuid);
                }
            };
        }));
        this.updateSubscribers = this.effect(tap((animations) => {
            if (!animations.size)
                return;
            const subscribers = Array.from(animations.values());
            subscribers.sort((a, b) => (a.priority || 0) - (b.priority || 0));
            this.set({
                hasPriority: subscribers.some(({ priority }) => !!priority),
                subscribers,
            });
        }));
        this.set({
            animations: new Map(),
            subscribers: [],
            hasPriority: false,
        });
    }
    init() {
        this.updateSubscribers(this.select((s) => s.animations));
    }
    register(animation) {
        const uuid = animation.obj instanceof THREE.Object3D
            ? animation.obj.uuid
            : makeId();
        this.registerAnimation(Object.assign(Object.assign({}, animation), { uuid }));
        return uuid;
    }
    unregister(uuid) {
        if (!uuid)
            return;
        const currentAnimations = this.get((s) => s.animations);
        const deleted = currentAnimations.delete(uuid);
        if (deleted) {
            this.set({ animations: currentAnimations });
        }
    }
}
NgtAnimationFrameStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAnimationFrameStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgtAnimationFrameStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAnimationFrameStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAnimationFrameStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return []; } });

class NgtLoop {
    constructor(canvasStore, animationFrameStore, zone) {
        this.canvasStore = canvasStore;
        this.animationFrameStore = animationFrameStore;
        this.zone = zone;
        this.running = false;
        this.frames = 0;
        this.loop = (timestamp) => this.zone.runOutsideAngular(() => {
            this.running = true;
            this.repeat = 0;
            const canvasState = this.canvasStore.get();
            if (canvasState.ready &&
                (canvasState.frameloop === 'always' || this.frames > 0)) {
                this.repeat += this.render(timestamp, canvasState, this.animationFrameStore.get());
            }
            if (this.repeat > 0)
                return requestAnimationFrame(this.loop);
            this.running = false;
            return;
        });
    }
    render(timestamp, canvasState, animationFrameState) {
        return this.zone.runOutsideAngular(() => {
            const { clock, frameloop, camera, scene, renderer, mouse, size, viewport, } = canvasState;
            const { subscribers, hasPriority } = animationFrameState;
            if (renderer) {
                let delta = clock.getDelta();
                // In frameloop='never' mode, clock times are updated using the provided timestamp
                if (frameloop === 'never' && typeof timestamp === 'number') {
                    delta = timestamp - clock.elapsedTime;
                    clock.oldTime = clock.elapsedTime;
                    clock.elapsedTime = timestamp;
                }
                const renderState = {
                    clock,
                    camera,
                    scene,
                    renderer,
                    mouse,
                    size,
                    viewport,
                    delta,
                };
                for (const subscriber of subscribers) {
                    const object = typeof subscriber.obj === 'function'
                        ? subscriber.obj()
                        : subscriber.obj;
                    subscriber.callback(renderState, object);
                }
                if (!hasPriority) {
                    renderer.render(scene, camera);
                }
            }
            this.frames = Math.max(0, this.frames - 1);
            return canvasState.frameloop === 'always' ? 1 : this.frames;
        });
    }
    invalidate(canvasState = this.canvasStore.get()) {
        var _a;
        if (canvasState.vr) {
            (_a = canvasState.renderer) === null || _a === void 0 ? void 0 : _a.setAnimationLoop((timestamp) => {
                this.render(timestamp, this.canvasStore.get(), this.animationFrameStore.get());
            });
            return;
        }
        if (!canvasState.ready || canvasState.frameloop === 'never')
            return;
        // Increase frames, do not go higher than 60
        this.frames = Math.min(60, this.frames + 1);
        // If the render-loop isn't active, start it
        if (!this.running) {
            this.running = true;
            requestAnimationFrame(this.loop);
        }
    }
}
NgtLoop.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoop, deps: [{ token: NgtCanvasStore }, { token: NgtAnimationFrameStore }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Injectable });
NgtLoop.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoop });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoop, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: NgtCanvasStore }, { type: NgtAnimationFrameStore }, { type: i0.NgZone }]; } });

class NgtPerformance {
    constructor(canvasStore, zone) {
        this.canvasStore = canvasStore;
        this.zone = zone;
        this.timeoutId = undefined;
    }
    regress() {
        this.zone.runOutsideAngular(() => {
            const { current, min, debounce } = this.canvasStore.get((s) => s.performance);
            // clear timeout
            if (this.timeoutId) {
                clearTimeout(this.timeoutId);
                this.timeoutId = undefined;
            }
            // set lower bound
            if (current !== min)
                this.canvasStore.set((state) => (Object.assign(Object.assign({}, state), { performance: Object.assign(Object.assign({}, state.performance), { current: min }) })));
            // go back to upper bound
            this.timeoutId = setTimeout(() => {
                // performanceState.current = performanceSnapshot.max;
                this.canvasStore.set((state) => (Object.assign(Object.assign({}, state), { performance: Object.assign(Object.assign({}, state.performance), { current: state.performance.max }) })));
            }, debounce);
        });
    }
}
NgtPerformance.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPerformance, deps: [{ token: NgtCanvasStore }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Injectable });
NgtPerformance.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPerformance });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPerformance, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: NgtCanvasStore }, { type: i0.NgZone }]; } });

/**
 * From r3f https://github.com/pmndrs/react-three-fiber/blob/master/packages/fiber/src/core/events.ts
 */
/** Release pointer captures.
 * This is called by releasePointerCapture in the API, and when an object is removed.
 */
function releaseInternalPointerCapture(capturedMap, obj, captures, pointerId) {
    const captureData = captures.get(obj);
    if (captureData) {
        captures.delete(obj);
        // If this was the last capturing object for this pointer
        if (captures.size === 0) {
            capturedMap.delete(pointerId);
            captureData.target.releasePointerCapture(pointerId);
        }
    }
}
function removeInteractivity(eventsStateGetter, object) {
    const { internal } = eventsStateGetter();
    // Removes every trace of an object from the data store
    internal.interaction = internal.interaction.filter((o) => o !== object);
    internal.initialHits = internal.initialHits.filter((o) => o !== object);
    internal.hovered.forEach((value, key) => {
        if (value.eventObject === object || value.object === object) {
            internal.hovered.delete(key);
        }
    });
    internal.capturedMap.forEach((captures, pointerId) => {
        releaseInternalPointerCapture(internal.capturedMap, object, captures, pointerId);
    });
}
function getRenderState(state) {
    const { mouse, camera, clock, renderer, scene, size, viewport } = state;
    const delta = clock.getDelta();
    return {
        size,
        mouse,
        camera: camera,
        delta,
        clock,
        renderer: renderer,
        scene: scene,
        viewport,
    };
}
function createEvents(stateGetter, eventsStateGetter) {
    const temp = new THREE.Vector3();
    /** Sets up defaultRaycaster */
    function prepareRay(event) {
        var _a, _b, _c, _d, _e;
        const state = stateGetter();
        const { raycaster, mouse, camera, size } = state;
        // https://github.com/pmndrs/react-three-fiber/pull/782
        // Events trigger outside of canvas when moved
        if (raycaster && camera) {
            const customOffsets = (_a = raycaster.computeOffsets) === null || _a === void 0 ? void 0 : _a.call(raycaster, event, getRenderState(state));
            const offsetX = (_b = customOffsets === null || customOffsets === void 0 ? void 0 : customOffsets.offsetX) !== null && _b !== void 0 ? _b : event.offsetX;
            const offsetY = (_c = customOffsets === null || customOffsets === void 0 ? void 0 : customOffsets.offsetY) !== null && _c !== void 0 ? _c : event.offsetY;
            const width = (_d = customOffsets === null || customOffsets === void 0 ? void 0 : customOffsets.width) !== null && _d !== void 0 ? _d : size.width;
            const height = (_e = customOffsets === null || customOffsets === void 0 ? void 0 : customOffsets.height) !== null && _e !== void 0 ? _e : size.height;
            mouse.set((offsetX / width) * 2 - 1, -(offsetY / height) * 2 + 1);
            raycaster.setFromCamera(mouse, camera);
        }
    }
    /** Calculates delta */
    function calculateDistance(event) {
        const { internal } = eventsStateGetter();
        const dx = event.offsetX - internal.initialClick[0];
        const dy = event.offsetY - internal.initialClick[1];
        return Math.round(Math.sqrt(dx * dx + dy * dy));
    }
    /** Returns true if an instance has a valid pointer-event registered, this excludes scroll, clicks etc */
    function filterPointerEvents(objects) {
        return objects.filter((obj) => ['move', 'over', 'enter', 'out', 'leave'].some((name) => { var _a, _b; return (_b = (_a = obj.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b[('pointer' + name)]; }));
    }
    function intersect(filter) {
        var _a;
        const canvasState = stateGetter();
        const { internal } = eventsStateGetter();
        const { raycaster } = canvasState;
        // Skip event handling when noEvents is set
        if (!(raycaster === null || raycaster === void 0 ? void 0 : raycaster.enabled))
            return [];
        const seen = new Set();
        const intersections = [];
        // Allow callers to eliminate event objects
        const eventsObjects = filter
            ? filter(internal.interaction)
            : internal.interaction;
        // Intersect known handler objects and filter against duplicates
        let intersects = raycaster
            .intersectObjects(eventsObjects, true)
            .filter((item) => {
            const id = makeId(item);
            if (seen.has(id))
                return false;
            seen.add(id);
            return true;
        });
        // https://github.com/mrdoob/three.js/issues/16031
        // Allow custom userland intersect sort order
        if (raycaster.filter)
            intersects = raycaster.filter(intersects, getRenderState(canvasState));
        for (const intersect of intersects) {
            let eventObject = intersect.object;
            // Bubble event up
            while (eventObject) {
                if ((_a = eventObject.__ngt) === null || _a === void 0 ? void 0 : _a.eventCount)
                    intersections.push(Object.assign(Object.assign({}, intersect), { eventObject }));
                eventObject = eventObject.parent;
            }
        }
        return intersections;
    }
    /**  Creates filtered intersects and returns an array of positive hits */
    function patchIntersects(intersections, event) {
        const { internal } = eventsStateGetter();
        // If the interaction is captured, make all capturing targets  part of the
        // intersect.
        if ('pointerId' in event && internal.capturedMap.has(event.pointerId)) {
            for (const captureData of internal.capturedMap
                .get(event.pointerId)
                .values()) {
                intersections.push(captureData.intersection);
            }
        }
        return intersections;
    }
    function cancelPointer(hits) {
        const { internal } = eventsStateGetter();
        Array.from(internal.hovered.values()).forEach((hoveredObj) => {
            var _a, _b;
            // When no objects were hit or the the hovered object wasn't found underneath the cursor
            // we call onPointerOut and delete the object from the hovered-elements map
            if (!hits.length ||
                !hits.find((hit) => hit.object === hoveredObj.object &&
                    hit.index === hoveredObj.index &&
                    hit.instanceId === hoveredObj.instanceId)) {
                const eventObject = hoveredObj.eventObject;
                const instance = eventObject.__ngt;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                internal.hovered.delete(makeId(hoveredObj));
                if (instance === null || instance === void 0 ? void 0 : instance.eventCount) {
                    // Clear out intersects, they are outdated by now
                    const data = Object.assign(Object.assign({}, hoveredObj), { intersections: hits || [] });
                    (_a = handlers === null || handlers === void 0 ? void 0 : handlers.pointerout) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                    (_b = handlers === null || handlers === void 0 ? void 0 : handlers.pointerleave) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                }
            }
        });
    }
    /**  Handles intersections by forwarding them to handlers */
    function handleIntersects(intersections, event, delta, callback) {
        const { raycaster, mouse, camera } = stateGetter();
        const { internal } = eventsStateGetter();
        // If anything has been found, forward it to the event listeners
        if (intersections.length && camera) {
            const unprojectedPoint = temp
                .set(mouse.x, mouse.y, 0)
                .unproject(camera);
            const localState = { stopped: false };
            for (const hit of intersections) {
                const hasPointerCapture = (id) => { var _a, _b; return (_b = (_a = internal.capturedMap.get(id)) === null || _a === void 0 ? void 0 : _a.has(hit.eventObject)) !== null && _b !== void 0 ? _b : false; };
                const setPointerCapture = (id) => {
                    const captureData = {
                        intersection: hit,
                        target: event.target,
                    };
                    if (internal.capturedMap.has(id)) {
                        // if the pointerId was previously captured, we add the hit to the
                        // event capturedMap.
                        internal.capturedMap
                            .get(id)
                            .set(hit.eventObject, captureData);
                    }
                    else {
                        // if the pointerId was not previously captured, we create a map
                        // containing the hitObject, and the hit. hitObject is used for
                        // faster access.
                        internal.capturedMap.set(id, new Map([[hit.eventObject, captureData]]));
                    }
                    // Call the original event now
                    event.target.setPointerCapture(id);
                };
                const releasePointerCapture = (id) => {
                    const captures = internal.capturedMap.get(id);
                    if (captures) {
                        releaseInternalPointerCapture(internal.capturedMap, hit.eventObject, captures, id);
                    }
                };
                // Add native event props
                const extractEventProps = {};
                // This iterates over the event's properties including the inherited ones. Native PointerEvents have most of their props as getters which are inherited, but polyfilled PointerEvents have them all as their own properties (i.e. not inherited). We can't use Object.keys() or Object.entries() as they only return "own" properties; nor Object.getPrototypeOf(event) as that *doesn't* return "own" properties, only inherited ones.
                for (const prop in event) {
                    const property = event[prop];
                    // Only copy over atomics, leave functions alone as these should be
                    // called as event.nativeEvent.fn()
                    if (typeof property !== 'function')
                        extractEventProps[prop] = property;
                }
                const raycastEvent = Object.assign(Object.assign(Object.assign({}, hit), extractEventProps), { spaceX: mouse.x, spaceY: mouse.y, intersections, stopped: localState.stopped, delta,
                    unprojectedPoint, ray: raycaster === null || raycaster === void 0 ? void 0 : raycaster.ray, camera: camera, 
                    // Hijack stopPropagation, which just sets a flag
                    stopPropagation: () => {
                        // https://github.com/pmndrs/react-three-fiber/issues/596
                        // Events are not allowed to stop propagation if the pointer has been captured
                        const capturesForPointer = 'pointerId' in event &&
                            internal.capturedMap.get(event.pointerId);
                        // We only authorize stopPropagation...
                        if (
                        // ...if this pointer hasn't been captured
                        !capturesForPointer ||
                            // ... or if the hit object is capturing the pointer
                            capturesForPointer.has(hit.eventObject)) {
                            raycastEvent.stopped = localState.stopped = true;
                            // Propagation is stopped, remove all other hover records
                            // An event handler is only allowed to flush other handlers if it is hovered itself
                            if (internal.hovered.size &&
                                Array.from(internal.hovered.values()).find((i) => i.eventObject === hit.eventObject)) {
                                // Objects cannot flush out higher up objects that have already caught the event
                                const higher = intersections.slice(0, intersections.indexOf(hit));
                                cancelPointer([...higher, hit]);
                            }
                        }
                    }, 
                    // there should be a distinction between target and currentTarget
                    target: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, currentTarget: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, sourceEvent: event, nativeEvent: event });
                // Call subscribers
                callback(raycastEvent);
                // Event bubbling may be interrupted by stopPropagation
                if (localState.stopped)
                    break;
            }
        }
        return intersections;
    }
    const handlePointer = (name) => {
        // Deal with cancelation
        switch (name) {
            case 'pointerleave':
            case 'pointercancel':
                return () => cancelPointer([]);
            case 'lostpointercapture':
                return (event) => {
                    const { internal: { capturedMap }, } = eventsStateGetter();
                    if ('pointerId' in event &&
                        !capturedMap.has(event.pointerId)) {
                        // If the object event interface had onLostPointerCapture, we'd call it here on every
                        // object that's getting removed.
                        capturedMap.delete(event.pointerId);
                        cancelPointer([]);
                    }
                };
        }
        // Any other pointer goes here ...
        return (event) => {
            const { pointerMissed, internal } = eventsStateGetter();
            prepareRay(event);
            // Get fresh intersects
            const isPointerMove = name === 'pointermove';
            const isClickEvent = name === 'click' ||
                name === 'contextmenu' ||
                name === 'dblclick';
            const filter = isPointerMove ? filterPointerEvents : undefined;
            const hits = patchIntersects(intersect(filter), event);
            const delta = isClickEvent ? calculateDistance(event) : 0;
            // Save initial coordinates on pointer-down
            if (name === 'pointerdown') {
                internal.initialClick = [event.offsetX, event.offsetY];
                internal.initialHits = hits.map((hit) => hit.eventObject);
            }
            // If a click yields no results, pass it back to the user as a miss
            // Missed events have to come first in order to establish user-land side-effect clean up
            if (isClickEvent && !hits.length) {
                if (delta <= 2) {
                    objectPointerMissed(event, internal.interaction);
                    if (pointerMissed)
                        pointerMissed(event);
                }
            }
            // Take care of unhover
            if (isPointerMove)
                cancelPointer(hits);
            handleIntersects(hits, event, delta, (data) => {
                var _a, _b, _c;
                const eventObject = data.eventObject;
                const instance = eventObject
                    .__ngt;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                // Check presence of handlers
                if (!(instance === null || instance === void 0 ? void 0 : instance.eventCount))
                    return;
                if (isPointerMove) {
                    // Move event ...
                    if ((handlers === null || handlers === void 0 ? void 0 : handlers.pointerover) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerenter) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerout) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerleave)) {
                        // When enter or out is present take care of hover-state
                        const id = makeId(data);
                        const hoveredItem = internal.hovered.get(id);
                        if (!hoveredItem) {
                            // If the object wasn't previously hovered, book it and call its handler
                            internal.hovered.set(id, data);
                            (_a = handlers.pointerover) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                            (_b = handlers.pointerenter) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                        }
                        else if (hoveredItem.stopped) {
                            // If the object was previously hovered and stopped, we shouldn't allow other items to proceed
                            data.stopPropagation();
                        }
                    }
                    // Call mouse move
                    (_c = handlers === null || handlers === void 0 ? void 0 : handlers.pointermove) === null || _c === void 0 ? void 0 : _c.call(handlers, data);
                }
                else {
                    // All other events ...
                    const handler = handlers === null || handlers === void 0 ? void 0 : handlers[name];
                    if (handler) {
                        // Forward all events back to their respective handlers with the exception of click events,
                        // which must use the initial target
                        if (!isClickEvent ||
                            internal.initialHits.includes(eventObject)) {
                            // Missed events have to come first
                            objectPointerMissed(event, internal.interaction.filter((object) => !internal.initialHits.includes(object)));
                            // Now call the handler
                            handler(data);
                        }
                    }
                }
            });
        };
    };
    function objectPointerMissed(event, objects) {
        objects.forEach((object) => { var _a, _b, _c; return (_c = (_b = (_a = object.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b.pointermissed) === null || _c === void 0 ? void 0 : _c.call(_b, event); });
    }
    return { handlePointer };
}

const names = {
    click: false,
    contextmenu: false,
    dblclick: false,
    wheel: false,
    pointerdown: true,
    pointerup: true,
    pointerleave: true,
    pointermove: true,
    pointercancel: true,
    lostpointercapture: true,
};
class NgtEventsStore extends NgtStore {
    constructor(canvasStore) {
        super();
        this.canvasStore = canvasStore;
        this.set({
            pointerMissed: noop,
            connected: false,
            handlers: {},
            internal: {
                interaction: [],
                hovered: new Map(),
                capturedMap: new Map(),
                initialClick: [0, 0],
                initialHits: [],
            },
        });
    }
    init(element) {
        const { handlePointer } = createEvents(this.canvasStore.get.bind(this.canvasStore), this.get.bind(this));
        this.set({
            handlers: Object.keys(names).reduce((handlers, supportedEventName) => {
                handlers[supportedEventName] =
                    handlePointer(supportedEventName);
                return handlers;
            }, {}),
        });
        this.connectElement(element);
    }
    addInteraction(interaction) {
        this.set((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: [...state.internal.interaction, interaction] }) })));
    }
    removeInteraction(uuid) {
        this.set((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: state.internal.interaction.filter((interaction) => interaction.uuid !== uuid) }) })));
    }
    connectElement(element) {
        this.set({ connected: element });
        const handlers = this.get((s) => s.handlers);
        Object.entries(handlers !== null && handlers !== void 0 ? handlers : {}).forEach(([eventName, handler]) => {
            const passive = names[eventName];
            element.addEventListener(eventName, handler, { passive });
        });
    }
    disconnectElement() {
        const { handlers, connected } = this.get();
        if (connected) {
            Object.entries(handlers !== null && handlers !== void 0 ? handlers : {}).forEach(([eventName, handler]) => {
                if (connected instanceof HTMLElement) {
                    connected.removeEventListener(eventName, handler);
                }
            });
        }
    }
    ngOnDestroy() {
        this.disconnectElement();
        super.ngOnDestroy();
    }
}
NgtEventsStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtEventsStore, deps: [{ token: NgtCanvasStore }], target: i0.ɵɵFactoryTarget.Injectable });
NgtEventsStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtEventsStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtEventsStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: NgtCanvasStore }]; } });

class NgtCanvas extends NgtStore {
    constructor(canvasStore, eventsStore, animationFrameStore, loop, zone, canvasOptions) {
        super();
        this.canvasStore = canvasStore;
        this.eventsStore = eventsStore;
        this.animationFrameStore = animationFrameStore;
        this.loop = loop;
        this.zone = zone;
        this.canvasOptions = canvasOptions;
        this.hostClass = true;
        this.created = new EventEmitter();
        this.pointermissed = new EventEmitter();
    }
    set vr(vr) {
        this.canvasStore.set({ vr: vr === '' ? true : vr });
    }
    set linear(linear) {
        this.canvasStore.set({ linear: linear === '' ? true : linear });
    }
    set flat(flat) {
        this.canvasStore.set({ flat: flat === '' ? true : flat });
    }
    set frameloop(frameloop) {
        this.canvasStore.set({ frameloop });
    }
    set orthographic(orthographic) {
        this.canvasStore.set({
            orthographic: orthographic === '' ? true : orthographic,
        });
    }
    set performance(performance) {
        this.canvasStore.set({ performance });
    }
    set size(size) {
        this.canvasStore.set({ size });
    }
    set dpr(dpr) {
        this.canvasStore.set({ dpr });
    }
    set clock(clock) {
        this.canvasStore.set({ clock });
    }
    set raycaster(raycaster) {
        this.canvasStore.set({ raycasterOptions: raycaster });
    }
    set shadows(shadows) {
        this.canvasStore.set({ shadows });
    }
    set camera(cameraOptions) {
        this.canvasStore.set({ cameraOptions });
    }
    set scene(sceneOptions) {
        this.canvasStore.set({ sceneOptions });
    }
    set gl(glOptions) {
        this.canvasStore.set({ glOptions });
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            // if there is handler to pointermissed on the canvas
            // update pointermissed in events store so that
            // events util will handle it
            if (this.pointermissed.observed) {
                this.eventsStore.set({
                    pointerMissed: (event) => {
                        this.pointermissed.emit(event);
                    },
                });
            }
            this.canvasStore.init(this.rendererCanvas.nativeElement);
            this.onCanvasReady(this.canvasStore.ready$, () => {
                const canvasState = this.canvasStore.get();
                this.eventsStore.init(canvasState.renderer.domElement);
                this.animationFrameStore.init();
                this.loop.invalidate(canvasState);
                this.created.emit(canvasState);
                if (this.canvasOptions.initialLog) {
                    console.group('Canvas initialized');
                    console.log(canvasState);
                    console.groupEnd();
                }
            });
        });
    }
}
NgtCanvas.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCanvas, deps: [{ token: NgtCanvasStore }, { token: NgtEventsStore }, { token: NgtAnimationFrameStore }, { token: NgtLoop }, { token: i0.NgZone }, { token: NGT_CANVAS_OPTIONS }], target: i0.ɵɵFactoryTarget.Component });
NgtCanvas.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "13.2.2", type: NgtCanvas, selector: "ngt-canvas", inputs: { vr: "vr", linear: "linear", flat: "flat", frameloop: "frameloop", orthographic: "orthographic", performance: "performance", size: "size", dpr: "dpr", clock: "clock", raycaster: "raycaster", shadows: "shadows", camera: "camera", scene: "scene", gl: "gl" }, outputs: { created: "created", pointermissed: "pointermissed" }, host: { properties: { "class.ngt-canvas": "this.hostClass" } }, providers: [
        NgtCanvasStore,
        NgtAnimationFrameStore,
        NgtEventsStore,
        NgtResize,
        NgtLoop,
        NgtPerformance,
    ], viewQueries: [{ propertyName: "rendererCanvas", first: true, predicate: ["rendererCanvas"], descendants: true, static: true }], usesInheritance: true, ngImport: i0, template: `
        <canvas #rendererCanvas></canvas>
        <ng-container
            *ngIf="canvasOptions.projectContent"
            [ngTemplateOutlet]="contentTemplate"
        >
        </ng-container>
        <ng-template #contentTemplate>
            <ng-content></ng-content>
        </ng-template>
    `, isInline: true, styles: [":host{display:block;position:relative;width:100%;height:100%;overflow:hidden}:host canvas{display:block}\n"], directives: [{ type: i5.NgIf, selector: "[ngIf]", inputs: ["ngIf", "ngIfThen", "ngIfElse"] }, { type: i5.NgTemplateOutlet, selector: "[ngTemplateOutlet]", inputs: ["ngTemplateOutletContext", "ngTemplateOutlet"] }], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCanvas, decorators: [{
            type: Component,
            args: [{
                    selector: 'ngt-canvas',
                    template: `
        <canvas #rendererCanvas></canvas>
        <ng-container
            *ngIf="canvasOptions.projectContent"
            [ngTemplateOutlet]="contentTemplate"
        >
        </ng-container>
        <ng-template #contentTemplate>
            <ng-content></ng-content>
        </ng-template>
    `,
                    styles: [
                        `
            :host {
                display: block;
                position: relative;
                width: 100%;
                height: 100%;
                overflow: hidden;
            }

            :host canvas {
                display: block;
            }
        `,
                    ],
                    changeDetection: ChangeDetectionStrategy.OnPush,
                    providers: [
                        NgtCanvasStore,
                        NgtAnimationFrameStore,
                        NgtEventsStore,
                        NgtResize,
                        NgtLoop,
                        NgtPerformance,
                    ],
                }]
        }], ctorParameters: function () {
        return [{ type: NgtCanvasStore }, { type: NgtEventsStore }, { type: NgtAnimationFrameStore }, { type: NgtLoop }, { type: i0.NgZone }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_CANVAS_OPTIONS]
                    }] }];
    }, propDecorators: { hostClass: [{
                type: HostBinding,
                args: ['class.ngt-canvas']
            }], vr: [{
                type: Input
            }], linear: [{
                type: Input
            }], flat: [{
                type: Input
            }], frameloop: [{
                type: Input
            }], orthographic: [{
                type: Input
            }], performance: [{
                type: Input
            }], size: [{
                type: Input
            }], dpr: [{
                type: Input
            }], clock: [{
                type: Input
            }], raycaster: [{
                type: Input
            }], shadows: [{
                type: Input
            }], camera: [{
                type: Input
            }], scene: [{
                type: Input
            }], gl: [{
                type: Input
            }], created: [{
                type: Output
            }], pointermissed: [{
                type: Output
            }], rendererCanvas: [{
                type: ViewChild,
                args: ['rendererCanvas', { static: true }]
            }] } });
class NgtCoreModule {
}
NgtCoreModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCoreModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCoreModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCoreModule, declarations: [NgtCanvas], imports: [CommonModule], exports: [NgtCanvas] });
NgtCoreModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCoreModule, imports: [[CommonModule]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCoreModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtCanvas],
                    exports: [NgtCanvas],
                    imports: [CommonModule],
                }]
        }] });

class NgtMathPipe {
    transform(value, keyOfMath) {
        const params = Array.isArray(value) ? value : [value];
        return Math[keyOfMath](...params);
    }
}
NgtMathPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtMathPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipe, name: "math" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'math' }]
        }] });
/**
 * @deprecated will be removed in v5.
 * This pipe is deprecated due to complicated template usage.
 * `Math.PI` is the most used one so see {@link NgtPiPipe} for Math.PI
 */
class NgtMathConstantPipe {
    /**
     * @deprecated will be removed in v5.
     * This pipe is deprecated due to complicated template usage.
     * `Math.PI` is the most used one so see {@link NgtPiPipe} for Math.PI
     */
    transform(value, keyOfMathConst) {
        if (keyOfMathConst === 'random') {
            return Math.random();
        }
        return value * Math[keyOfMathConst];
    }
}
NgtMathConstantPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathConstantPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtMathConstantPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathConstantPipe, name: "mathConst" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathConstantPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'mathConst' }]
        }] });
class NgtMathPipeModule {
}
NgtMathPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtMathPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipeModule, declarations: [NgtMathPipe, NgtMathConstantPipe], exports: [NgtMathPipe, NgtMathConstantPipe] });
NgtMathPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMathPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtMathPipe, NgtMathConstantPipe],
                    exports: [NgtMathPipe, NgtMathConstantPipe],
                }]
        }] });

class NgtVector2Pipe {
    transform(value) {
        return makeVector2(value);
    }
}
NgtVector2Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector2Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector2Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector2Pipe, name: "vector2" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector2Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector2' }]
        }] });
class NgtVector3Pipe {
    transform(value) {
        return makeVector3(value);
    }
}
NgtVector3Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector3Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector3Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector3Pipe, name: "vector3" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector3Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector3' }]
        }] });
class NgtVector4Pipe {
    transform(value) {
        return makeVector4(value);
    }
}
NgtVector4Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector4Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtVector4Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector4Pipe, name: "vector4" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVector4Pipe, decorators: [{
            type: Pipe,
            args: [{ name: 'vector4' }]
        }] });
class NgtVectorPipeModule {
}
NgtVectorPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVectorPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtVectorPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVectorPipeModule, declarations: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe], exports: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe] });
NgtVectorPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVectorPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtVectorPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe],
                    exports: [NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe],
                }]
        }] });

class NgtColorPipe {
    /**
     * ConstructorParameters<typeof THREE.Color> has a limitation on THREE.Color constructor overloads
     */
    transform(args) {
        return makeColor(args);
    }
}
NgtColorPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtColorPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipe, name: "color" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'color' }]
        }] });
class NgtColorPipeModule {
}
NgtColorPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtColorPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipeModule, declarations: [NgtColorPipe], exports: [NgtColorPipe] });
NgtColorPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtColorPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtColorPipe],
                    exports: [NgtColorPipe],
                }]
        }] });

class NgtFogPipe {
    transform(value) {
        return new THREE.Fog(...value);
    }
}
NgtFogPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtFogPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipe, name: "fog" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'fog',
                }]
        }] });
class NgtFogExp2Pipe {
    transform(value) {
        return new THREE.FogExp2(...value);
    }
}
NgtFogExp2Pipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogExp2Pipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtFogExp2Pipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogExp2Pipe, name: "fogExp2" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogExp2Pipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'fogExp2',
                }]
        }] });
class NgtFogPipeModule {
}
NgtFogPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtFogPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipeModule, declarations: [NgtFogPipe, NgtFogExp2Pipe], exports: [NgtFogPipe, NgtFogExp2Pipe] });
NgtFogPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtFogPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtFogPipe, NgtFogExp2Pipe],
                    exports: [NgtFogPipe, NgtFogExp2Pipe],
                }]
        }] });

class NgtSidePipe {
    transform(side) {
        switch (side) {
            case 'front':
                return THREE.FrontSide;
            case 'back':
                return THREE.BackSide;
            case 'double':
                return THREE.DoubleSide;
        }
    }
}
NgtSidePipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtSidePipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipe, name: "side" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipe, decorators: [{
            type: Pipe,
            args: [{ name: 'side' }]
        }] });
class NgtSidePipeModule {
}
NgtSidePipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtSidePipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipeModule, declarations: [NgtSidePipe], exports: [NgtSidePipe] });
NgtSidePipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtSidePipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtSidePipe],
                    exports: [NgtSidePipe],
                }]
        }] });

class NgtPiPipe {
    transform(value) {
        return value * Math.PI;
    }
}
NgtPiPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtPiPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipe, name: "pi" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'pi' }]
        }] });
class NgtPiPipeModule {
}
NgtPiPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtPiPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipeModule, declarations: [NgtPiPipe], exports: [NgtPiPipe] });
NgtPiPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtPiPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtPiPipe],
                    exports: [NgtPiPipe],
                }]
        }] });

class NgtRadianPipe {
    transform(degree) {
        return (degree * Math.PI) / 180;
    }
}
NgtRadianPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtRadianPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipe, name: "radian" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipe, decorators: [{
            type: Pipe,
            args: [{ name: 'radian' }]
        }] });
class NgtRadianPipeModule {
}
NgtRadianPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtRadianPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipeModule, declarations: [NgtRadianPipe], exports: [NgtRadianPipe] });
NgtRadianPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRadianPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtRadianPipe],
                    exports: [NgtRadianPipe],
                }]
        }] });

class NgtRepeat extends NgForOf {
    set ngForRepeat(count) {
        this.ngForOf = Number.isInteger(count)
            ? Array.from({ length: count }, (_, i) => i)
            : [];
    }
}
NgtRepeat.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeat, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtRepeat.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtRepeat, selector: "[ngFor][ngForRepeat]", inputs: { ngForRepeat: "ngForRepeat" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeat, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngFor][ngForRepeat]',
                }]
        }], propDecorators: { ngForRepeat: [{
                type: Input
            }] } });
class NgtRepeatModule {
}
NgtRepeatModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeatModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtRepeatModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeatModule, declarations: [NgtRepeat], exports: [NgtRepeat] });
NgtRepeatModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeatModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtRepeatModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtRepeat],
                    exports: [NgtRepeat],
                }]
        }] });

class NgtDestroyed extends Subject {
    ngOnDestroy() {
        this.next();
        this.complete();
    }
}
NgtDestroyed.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtDestroyed, deps: null, target: i0.ɵɵFactoryTarget.Injectable });
NgtDestroyed.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtDestroyed });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtDestroyed, decorators: [{
            type: Injectable
        }] });

class Controller {
    constructor() {
        this.disabled = false;
        this.changes$ = new ReplaySubject(1);
    }
    ngOnChanges(changes) {
        this.changes$.next(changes);
    }
}
Controller.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: Controller, deps: [], target: i0.ɵɵFactoryTarget.Directive });
Controller.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: Controller, inputs: { disabled: "disabled" }, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: Controller, decorators: [{
            type: Directive
        }], propDecorators: { disabled: [{
                type: Input
            }] } });
function controllerFactory(newInstanceOnNull = false, controllerType) {
    return (controller, cdr, destroyed) => {
        if (!controller) {
            return newInstanceOnNull ? new controllerType() : null;
        }
        controller.changes$.pipe(takeUntil(destroyed)).subscribe(() => {
            cdr.markForCheck();
        });
        return controller;
    };
}
function createControllerProviderFactory({ watchedControllerTokenName, controller, newInstanceOnNull = false, }) {
    const watchedControllerToken = new InjectionToken(watchedControllerTokenName);
    const controllerProvider = [
        NgtDestroyed,
        {
            provide: watchedControllerToken,
            deps: [
                [new Optional(), controller],
                ChangeDetectorRef,
                NgtDestroyed,
            ],
            useFactory: controllerFactory(newInstanceOnNull, controller),
        },
    ];
    return [watchedControllerToken, controllerProvider];
}

// GENERATED
class NgtObjectInputsController extends Controller {
    constructor() {
        super(...arguments);
        this.visible = true;
        this.matrixAutoUpdate = true;
        this.appendMode = 'immediate';
        // events
        this.click = new EventEmitter();
        this.contextmenu = new EventEmitter();
        this.dblclick = new EventEmitter();
        this.pointerup = new EventEmitter();
        this.pointerdown = new EventEmitter();
        this.pointerover = new EventEmitter();
        this.pointerout = new EventEmitter();
        this.pointerenter = new EventEmitter();
        this.pointerleave = new EventEmitter();
        this.pointermove = new EventEmitter();
        this.pointermissed = new EventEmitter();
        this.pointercancel = new EventEmitter();
        this.wheel = new EventEmitter();
    }
    set position(position) {
        this._position = makeVector3(position);
    }
    get position() {
        return this._position;
    }
    set rotation(rotation) {
        this._rotation = makeForSet(THREE.Euler, rotation);
    }
    get rotation() {
        return this._rotation;
    }
    set quaternion(quaternion) {
        this._quaternion = makeForSet(THREE.Quaternion, quaternion);
    }
    get quaternion() {
        return this._quaternion;
    }
    set scale(scale) {
        this._scale = makeVector3(scale);
    }
    get scale() {
        return this._scale;
    }
    set color(color) {
        this._color = makeColor(color);
    }
    get color() {
        return this._color;
    }
    get object3dProps() {
        var _a, _b, _c, _d;
        return {
            name: this.name,
            position: (_a = this._position) === null || _a === void 0 ? void 0 : _a.toArray(),
            rotation: (_b = this._rotation) === null || _b === void 0 ? void 0 : _b.toArray(),
            quaternion: (_c = this._quaternion) === null || _c === void 0 ? void 0 : _c.toArray(),
            scale: (_d = this._scale) === null || _d === void 0 ? void 0 : _d.toArray(),
            color: this.color,
            userData: this.userData,
            dispose: this.dispose,
            raycast: this.raycast,
            castShadow: this.castShadow,
            receiveShadow: this.receiveShadow,
            visible: this.visible,
            matrixAutoUpdate: this.matrixAutoUpdate,
        };
    }
}
NgtObjectInputsController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsController, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtObjectInputsController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtObjectInputsController, selector: "\n    ngt-primitive,\n    ngt-bone,\n    ngt-group,\n    ngt-lod,\n    ngt-points,\n    ngt-mesh,\n    ngt-instanced-mesh,\n    ngt-skinned-mesh,\n    ngt-audio,\n    ngt-positional-audio,\n    ngt-line,\n    ngt-line-loop,\n    ngt-line-segments,\n    ngt-light-probe,\n    ngt-ambient-light,\n    ngt-ambient-light-probe,\n    ngt-hemisphere-light,\n    ngt-hemisphere-light-probe,\n    ngt-directional-light,\n    ngt-point-light,\n    ngt-spot-light,\n    ngt-rect-area-light,\n    ngt-arrow-helper,\n    ngt-axes-helper,\n    ngt-box3-helper,\n    ngt-grid-helper,\n    ngt-plane-helper,\n    ngt-polar-grid-helper,\n    ngt-sprite,\n    ngt-camera,\n    ngt-perspective-camera,\n    ngt-orthographic-camera,\n    ngt-array-camera,\n    ngt-stereo-camera,\n    ngt-cube-camera,\n    ngt-soba-billboard,\n    ngt-soba-detailed,\n    ngt-soba-line,\n    ngt-soba-quadratic-bezier-line,\n    ngt-soba-cubic-bezier-line,\n    ngt-soba-orthographic-camera,\n    ngt-soba-gizmo-helper,\n    ngt-soba-gizmo-viewport,\n    ngt-soba-gizmo-axis-head,\n    ngt-soba-transform-controls,\n    ngt-soba-center,\n    ngt-soba-bounds,\n    ngt-soba-float,\n    ngt-soba-stage,\n    ngt-soba-backdrop,\n    ngt-soba-spot-light,\n    ngt-soba-html,\n    ngt-soba-contact-shadows,\n    ngt-soba-text\n  ", inputs: { name: "name", position: "position", rotation: "rotation", quaternion: "quaternion", scale: "scale", color: "color", userData: "userData", castShadow: "castShadow", receiveShadow: "receiveShadow", visible: "visible", matrixAutoUpdate: "matrixAutoUpdate", dispose: "dispose", raycast: "raycast", appendMode: "appendMode", appendTo: "appendTo" }, outputs: { click: "click", contextmenu: "contextmenu", dblclick: "dblclick", pointerup: "pointerup", pointerdown: "pointerdown", pointerover: "pointerover", pointerout: "pointerout", pointerenter: "pointerenter", pointerleave: "pointerleave", pointermove: "pointermove", pointermissed: "pointermissed", pointercancel: "pointercancel", wheel: "wheel" }, exportAs: ["ngtObjectInputsController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-primitive,
    ngt-bone,
    ngt-group,
    ngt-lod,
    ngt-points,
    ngt-mesh,
    ngt-instanced-mesh,
    ngt-skinned-mesh,
    ngt-audio,
    ngt-positional-audio,
    ngt-line,
    ngt-line-loop,
    ngt-line-segments,
    ngt-light-probe,
    ngt-ambient-light,
    ngt-ambient-light-probe,
    ngt-hemisphere-light,
    ngt-hemisphere-light-probe,
    ngt-directional-light,
    ngt-point-light,
    ngt-spot-light,
    ngt-rect-area-light,
    ngt-arrow-helper,
    ngt-axes-helper,
    ngt-box3-helper,
    ngt-grid-helper,
    ngt-plane-helper,
    ngt-polar-grid-helper,
    ngt-sprite,
    ngt-camera,
    ngt-perspective-camera,
    ngt-orthographic-camera,
    ngt-array-camera,
    ngt-stereo-camera,
    ngt-cube-camera,
    ngt-soba-billboard,
    ngt-soba-detailed,
    ngt-soba-line,
    ngt-soba-quadratic-bezier-line,
    ngt-soba-cubic-bezier-line,
    ngt-soba-orthographic-camera,
    ngt-soba-gizmo-helper,
    ngt-soba-gizmo-viewport,
    ngt-soba-gizmo-axis-head,
    ngt-soba-transform-controls,
    ngt-soba-center,
    ngt-soba-bounds,
    ngt-soba-float,
    ngt-soba-stage,
    ngt-soba-backdrop,
    ngt-soba-spot-light,
    ngt-soba-html,
    ngt-soba-contact-shadows,
    ngt-soba-text
  `,
                    exportAs: 'ngtObjectInputsController',
                }]
        }], propDecorators: { name: [{
                type: Input
            }], position: [{
                type: Input
            }], rotation: [{
                type: Input
            }], quaternion: [{
                type: Input
            }], scale: [{
                type: Input
            }], color: [{
                type: Input
            }], userData: [{
                type: Input
            }], castShadow: [{
                type: Input
            }], receiveShadow: [{
                type: Input
            }], visible: [{
                type: Input
            }], matrixAutoUpdate: [{
                type: Input
            }], dispose: [{
                type: Input
            }], raycast: [{
                type: Input
            }], appendMode: [{
                type: Input
            }], appendTo: [{
                type: Input
            }], click: [{
                type: Output
            }], contextmenu: [{
                type: Output
            }], dblclick: [{
                type: Output
            }], pointerup: [{
                type: Output
            }], pointerdown: [{
                type: Output
            }], pointerover: [{
                type: Output
            }], pointerout: [{
                type: Output
            }], pointerenter: [{
                type: Output
            }], pointerleave: [{
                type: Output
            }], pointermove: [{
                type: Output
            }], pointermissed: [{
                type: Output
            }], pointercancel: [{
                type: Output
            }], wheel: [{
                type: Output
            }] } });
class NgtObjectInputsControllerModule {
}
NgtObjectInputsControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtObjectInputsControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsControllerModule, declarations: [NgtObjectInputsController], exports: [NgtObjectInputsController] });
NgtObjectInputsControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsControllerModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectInputsControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtObjectInputsController],
                    exports: [NgtObjectInputsController],
                }]
        }] });
const [NGT_OBJECT_INPUTS_WATCHED_CONTROLLER, NGT_OBJECT_INPUTS_CONTROLLER_PROVIDER,] = createControllerProviderFactory({
    watchedControllerTokenName: 'Watched ObjectInputsController',
    controller: NgtObjectInputsController,
});

class NgtCursor {
    constructor(objectInputsController, document, destroyed) {
        if (!objectInputsController)
            return;
        merge(objectInputsController.pointerover.pipe(mapTo(true)), objectInputsController.pointerout.pipe(mapTo(false)))
            .pipe(takeUntil(destroyed))
            .subscribe((hovered) => {
            document.body.style.cursor = hovered ? 'pointer' : 'auto';
        });
    }
}
NgtCursor.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursor, deps: [{ token: NgtObjectInputsController, optional: true, self: true }, { token: DOCUMENT }, { token: NgtDestroyed }], target: i0.ɵɵFactoryTarget.Directive });
NgtCursor.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCursor, selector: "[ngtCursor]", providers: [NgtDestroyed], exportAs: ["ngtCursor"], ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursor, decorators: [{
            type: Directive,
            args: [{
                    selector: '[ngtCursor]',
                    exportAs: 'ngtCursor',
                    providers: [NgtDestroyed],
                }]
        }], ctorParameters: function () {
        return [{ type: NgtObjectInputsController, decorators: [{
                        type: Optional
                    }, {
                        type: Self
                    }] }, { type: Document, decorators: [{
                        type: Inject,
                        args: [DOCUMENT]
                    }] }, { type: NgtDestroyed }];
    } });
class NgtCursorModule {
}
NgtCursorModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursorModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCursorModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursorModule, declarations: [NgtCursor], exports: [NgtCursor] });
NgtCursorModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursorModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCursorModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtCursor],
                    exports: [NgtCursor],
                }]
        }] });

const supportedEvents = [
    'click',
    'contextmenu',
    'dblclick',
    'pointerup',
    'pointerdown',
    'pointerover',
    'pointerout',
    'pointerenter',
    'pointerleave',
    'pointermove',
    'pointermissed',
    'pointercancel',
    'wheel',
];
class NgtObjectController extends Controller {
    constructor(store, zone, canvasStore, eventsStore, animationFrameStore, objectInputsController, parentObjectFn, hostParentObjectFn) {
        super();
        this.store = store;
        this.zone = zone;
        this.canvasStore = canvasStore;
        this.eventsStore = eventsStore;
        this.animationFrameStore = animationFrameStore;
        this.objectInputsController = objectInputsController;
        this.parentObjectFn = parentObjectFn;
        this.hostParentObjectFn = hostParentObjectFn;
        this.priority = 0;
        this.useHostParent = false;
        this.animateReady = new EventEmitter();
        this.$object = new BehaviorSubject(null);
        this.inputChanges$ = defer(() => {
            return this.objectInputsController.changes$.pipe(map((changes) => {
                return Object.keys(changes).reduce((inputChanges, key) => {
                    inputChanges[key] =
                        this.objectInputsController[key];
                    return inputChanges;
                }, {});
            }));
        });
    }
    get object() {
        return this.$object.getValue();
    }
    set initFn(v) {
        this._initFn = v;
    }
    get initFn() {
        if (!this._initFn) {
            this.initFn = () => this.object;
        }
        return this._initFn;
    }
    ngOnInit() {
        this.store.effect(tap((changes) => {
            if (this.object) {
                this.applyCustomProps(changes);
            }
        }))(this.inputChanges$);
    }
    init() {
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            this.initSubscription = this.store.onCanvasReady(this.canvasStore.ready$, () => {
                if (this.object) {
                    this.switch();
                }
                else {
                    this.$object.next(this.initFn());
                }
                if (this.object) {
                    this.inputChanges$
                        .pipe(take(1))
                        .subscribe((changes) => {
                        this.applyCustomProps(changes);
                    });
                    if (!this.disabled) {
                        const observedEvents = supportedEvents.reduce((result, event) => {
                            const controllerEvent = this
                                .objectInputsController[event].observed
                                ? this.objectInputsController[event]
                                : null;
                            if (controllerEvent) {
                                result.handlers[event] =
                                    this.eventNameToHandler(controllerEvent);
                                result.eventCount += 1;
                            }
                            return result;
                        }, { handlers: {}, eventCount: 0 });
                        // setup __ngt instance
                        applyProps(this.object, {
                            __ngt: {
                                stateGetter: this.canvasStore.get.bind(this.canvasStore),
                                eventsStateGetter: this.eventsStore.get.bind(this.eventsStore),
                                handlers: observedEvents.handlers,
                                eventCount: observedEvents.eventCount,
                                linear: this.canvasStore.isLinear,
                            },
                        });
                        // add as an interaction if there are events observed
                        if (observedEvents.eventCount > 0) {
                            this.eventsStore.addInteraction(this.object);
                        }
                        this.canvasStore.set((state) => (Object.assign(Object.assign({}, state), { objects: Object.assign(Object.assign({}, state.objects), { [this.object.uuid]: this.object }) })));
                    }
                    if (this.objectInputsController.appendMode !== 'none' &&
                        this.object.isObject3D) {
                        this.appendToParent();
                    }
                    if (this.readyFn) {
                        this.readyFn();
                    }
                    if (this.animateReady.observed) {
                        this.animationFrameStore.register({
                            obj: () => this.object,
                            callback: (state) => {
                                this.animateReady.emit({
                                    state,
                                    object: this.object,
                                });
                            },
                            priority: this.priority,
                        });
                    }
                }
            });
        });
    }
    ngOnDestroy() {
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            if (this.object) {
                this.animationFrameStore.unregister(this.object.uuid);
                this.remove();
                if (this.object.clear) {
                    this.object.clear();
                }
                this.canvasStore.set((state) => {
                    const _a = state.objects, _b = this.object.uuid, _ = _a[_b], objects = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
                    return Object.assign(Object.assign({}, state), { objects });
                });
                this.eventsStore.removeInteraction(this.object.uuid);
            }
        });
    }
    appendToParent() {
        requestAnimationFrame(() => {
            if (this.objectInputsController.appendTo) {
                this.appendTo.add(this.object);
                return;
            }
            if (this.objectInputsController.appendMode === 'root') {
                this.addToScene();
                return;
            }
            if (this.objectInputsController.appendMode === 'immediate') {
                this.addToParent();
            }
        });
    }
    addToScene() {
        const scene = this.canvasStore.get((s) => s.scene);
        if (scene) {
            scene.add(this.object);
        }
    }
    addToParent() {
        var _a, _b, _c;
        let parentObject = (((_a = this.hostParentObjectFn) === null || _a === void 0 ? void 0 : _a.call(this)) ||
            ((_b = this.parentObjectFn) === null || _b === void 0 ? void 0 : _b.call(this)));
        if (this.useHostParent) {
            parentObject = (_c = this.hostParentObjectFn) === null || _c === void 0 ? void 0 : _c.call(this);
        }
        if (parentObject && parentObject.uuid !== this.object.uuid) {
            parentObject.add(this.object);
        }
        else {
            this.addToScene();
        }
    }
    eventNameToHandler(controllerEvent) {
        return (event) => {
            this.zone.run(() => {
                controllerEvent.emit(event);
            });
        };
    }
    switch() {
        const newObject3d = this.initFn();
        if (this.object.children) {
            this.object.traverse((object) => {
                if (object !== this.object && object.parent === this.object) {
                    object.parent = newObject3d;
                }
            });
            this.object.children = [];
        }
        this.remove();
        this.canvasStore.set((state) => {
            const _a = state.objects, _b = this.object.uuid, _ = _a[_b], objects = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { objects });
        });
        this.eventsStore.removeInteraction(this.object.uuid);
        this.$object.next(newObject3d);
    }
    remove() {
        var _a, _b;
        if (this.objectInputsController.appendTo) {
            if (this.appendTo) {
                this.appendTo.remove(this.object);
            }
        }
        else if (((_a = this.parentObjectFn) === null || _a === void 0 ? void 0 : _a.call(this)) &&
            ((_b = this.parentObjectFn()) === null || _b === void 0 ? void 0 : _b.uuid) !== this.object.uuid &&
            this.objectInputsController.appendMode === 'immediate') {
            this.parentObjectFn().remove(this.object);
        }
        else {
            const scene = this.canvasStore.get((s) => s.scene);
            if (scene) {
                scene.remove(this.object);
            }
        }
    }
    get appendTo() {
        return typeof this.objectInputsController.appendTo === 'function'
            ? this.objectInputsController.appendTo()
            : this.objectInputsController.appendTo;
    }
    applyCustomProps(changes) {
        this.zone.runOutsideAngular(() => {
            var _a, _b;
            const customProps = {
                castShadow: this.objectInputsController.castShadow,
                receiveShadow: this.objectInputsController.receiveShadow,
                visible: this.objectInputsController.visible,
                matrixAutoUpdate: this.objectInputsController.matrixAutoUpdate,
            };
            if (this.objectInputsController.name) {
                customProps['name'] = this.objectInputsController.name;
            }
            if (this.objectInputsController.position) {
                customProps['position'] = this.objectInputsController.position;
            }
            if (this.objectInputsController.rotation) {
                customProps['rotation'] = this.objectInputsController.rotation;
            }
            else if (this.objectInputsController.quaternion) {
                customProps['quaternion'] =
                    this.objectInputsController.quaternion;
            }
            if (this.objectInputsController.scale) {
                customProps['scale'] = this.objectInputsController.scale;
            }
            if (this.objectInputsController.userData) {
                customProps['userData'] = this.objectInputsController.userData;
            }
            if (this.objectInputsController.color) {
                customProps['color'] = this.objectInputsController.color;
                if (!this.canvasStore.isLinear) {
                    customProps['color'].convertSRGBToLinear();
                }
            }
            if (this.objectInputsController.dispose) {
                customProps['dispose'] = this.objectInputsController.dispose;
            }
            if (this.objectInputsController.raycast) {
                customProps['raycast'] = this.objectInputsController.raycast;
            }
            if (changes) {
                Object.assign(customProps, changes);
            }
            this.inputChanges$.pipe(take(1)).subscribe((changes) => {
                if (changes && Object.keys(changes).length) {
                    for (const [inputName, inputChange] of Object.entries(changes)) {
                        // TODO: double check this
                        if ([
                            'name',
                            'position',
                            'rotation',
                            'quaternion',
                            'scale',
                            'userData',
                            'color',
                            'dispose',
                            'raycast',
                            'castShadow',
                            'receiveShadow',
                            'visible',
                            'matrixAutoUpdate',
                        ].includes(inputName) // skip 14 common inputs
                        ) {
                            continue;
                        }
                        customProps[inputName] = inputChange;
                    }
                }
            });
            applyProps(this.object, customProps);
            (_b = (_a = this.object).updateMatrix) === null || _b === void 0 ? void 0 : _b.call(_a);
        });
    }
}
NgtObjectController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectController, deps: [{ token: NgtStore }, { token: i0.NgZone }, { token: NgtCanvasStore }, { token: NgtEventsStore }, { token: NgtAnimationFrameStore }, { token: NGT_OBJECT_INPUTS_WATCHED_CONTROLLER }, { token: NGT_PARENT_OBJECT, optional: true, skipSelf: true }, { token: NGT_HOST_PARENT_OBJECT, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtObjectController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtObjectController, selector: "\n        ngt-primitive,\n        ngt-bone,\n        ngt-group,\n        ngt-lod,\n        ngt-points,\n        ngt-mesh,\n        ngt-instanced-mesh,\n        ngt-skinned-mesh,\n        ngt-audio,\n        ngt-positional-audio,\n        ngt-line,\n        ngt-line-loop,\n        ngt-line-segments,\n        ngt-light-probe,\n        ngt-ambient-light,\n        ngt-ambient-light-probe,\n        ngt-hemisphere-light,\n        ngt-hemisphere-light-probe,\n        ngt-directional-light,\n        ngt-point-light,\n        ngt-spot-light,\n        ngt-rect-area-light,\n        ngt-arrow-helper,\n        ngt-axes-helper,\n        ngt-box3-helper,\n        ngt-grid-helper,\n        ngt-plane-helper,\n        ngt-polar-grid-helper,\n        ngt-sprite,\n        ngt-camera,\n        ngt-perspective-camera,\n        ngt-orthographic-camera,\n        ngt-array-camera,\n        ngt-stereo-camera,\n        ngt-cube-camera\n    ", inputs: { priority: "priority", useHostParent: "useHostParent" }, outputs: { animateReady: "animateReady" }, providers: [NGT_OBJECT_INPUTS_CONTROLLER_PROVIDER, NgtStore], exportAs: ["ngtObjectController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
        ngt-primitive,
        ngt-bone,
        ngt-group,
        ngt-lod,
        ngt-points,
        ngt-mesh,
        ngt-instanced-mesh,
        ngt-skinned-mesh,
        ngt-audio,
        ngt-positional-audio,
        ngt-line,
        ngt-line-loop,
        ngt-line-segments,
        ngt-light-probe,
        ngt-ambient-light,
        ngt-ambient-light-probe,
        ngt-hemisphere-light,
        ngt-hemisphere-light-probe,
        ngt-directional-light,
        ngt-point-light,
        ngt-spot-light,
        ngt-rect-area-light,
        ngt-arrow-helper,
        ngt-axes-helper,
        ngt-box3-helper,
        ngt-grid-helper,
        ngt-plane-helper,
        ngt-polar-grid-helper,
        ngt-sprite,
        ngt-camera,
        ngt-perspective-camera,
        ngt-orthographic-camera,
        ngt-array-camera,
        ngt-stereo-camera,
        ngt-cube-camera
    `,
                    exportAs: 'ngtObjectController',
                    providers: [NGT_OBJECT_INPUTS_CONTROLLER_PROVIDER, NgtStore],
                }]
        }], ctorParameters: function () {
        return [{ type: NgtStore }, { type: i0.NgZone }, { type: NgtCanvasStore }, { type: NgtEventsStore }, { type: NgtAnimationFrameStore }, { type: NgtObjectInputsController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_INPUTS_WATCHED_CONTROLLER]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_PARENT_OBJECT]
                    }] }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: SkipSelf
                    }, {
                        type: Inject,
                        args: [NGT_HOST_PARENT_OBJECT]
                    }] }];
    }, propDecorators: { priority: [{
                type: Input
            }], useHostParent: [{
                type: Input
            }], animateReady: [{
                type: Output
            }] } });
class NgtObjectControllerModule {
}
NgtObjectControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtObjectControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectControllerModule, declarations: [NgtObjectController], exports: [NgtObjectController, NgtObjectInputsControllerModule] });
NgtObjectControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectControllerModule, imports: [NgtObjectInputsControllerModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtObjectController],
                    exports: [NgtObjectController, NgtObjectInputsControllerModule],
                }]
        }] });
const [NGT_OBJECT_WATCHED_CONTROLLER, NGT_OBJECT_CONTROLLER_PROVIDER] = createControllerProviderFactory({
    watchedControllerTokenName: 'Watched ObjectController',
    controller: NgtObjectController,
});

class NgtGeometry extends NgtStore {
    constructor(zone, canvasStore, parentObjectFactory) {
        super();
        this.zone = zone;
        this.canvasStore = canvasStore;
        this.parentObjectFactory = parentObjectFactory;
        this.ready = this.select((s) => s.geometry);
        this.init = this.effect(tapEffect((geometryArgs) => {
            var _a;
            const geometry = new this.geometryType(...geometryArgs);
            const parentObject = (_a = this.parentObjectFactory) === null || _a === void 0 ? void 0 : _a.call(this);
            if (parentObject) {
                parentObject.geometry = geometry;
            }
            this.set({ geometry });
            return () => {
                geometry.dispose();
            };
        }));
        this.set({ geometryArgs: [] });
    }
    set geometryArgs(v) {
        this.set({ geometryArgs: Array.isArray(v) ? v : [v] });
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.canvasStore.ready$, () => {
                this.init(this.select((s) => s.geometryArgs));
            });
        });
    }
    get geometry() {
        return this.get((s) => s.geometry);
    }
}
NgtGeometry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtGeometry, deps: [{ token: i0.NgZone }, { token: NgtCanvasStore }, { token: NGT_PARENT_OBJECT, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtGeometry.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtGeometry, outputs: { ready: "ready" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtGeometry, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtCanvasStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_PARENT_OBJECT]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }] } });

// GENERATED
class NgtWithGeometryController extends Controller {
    constructor(zone) {
        super();
        this.zone = zone;
        this._geometry = undefined;
    }
    set geometry(v) {
        this._geometry = v;
    }
    get geometry() {
        return this._geometry;
    }
    set contentGeometry(dir) {
        if (dir) {
            this.zone.runOutsideAngular(() => {
                requestAnimationFrame(() => {
                    this.geometry = dir.geometry;
                });
            });
        }
    }
}
NgtWithGeometryController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryController, deps: [{ token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtWithGeometryController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtWithGeometryController, selector: "\n    ngt-mesh,\n    ngt-instanced-mesh,\n    ngt-skinned-mesh,\n    ngt-line,\n    ngt-line-loop,\n    ngt-line-segments,\n    ngt-points\n  ", inputs: { geometry: "geometry" }, queries: [{ propertyName: "contentGeometry", first: true, predicate: NgtGeometry, descendants: true, static: true }], exportAs: ["ngtWithGeometryController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-mesh,
    ngt-instanced-mesh,
    ngt-skinned-mesh,
    ngt-line,
    ngt-line-loop,
    ngt-line-segments,
    ngt-points
  `,
                    exportAs: 'ngtWithGeometryController',
                }]
        }], ctorParameters: function () { return [{ type: i0.NgZone }]; }, propDecorators: { geometry: [{
                type: Input
            }], contentGeometry: [{
                type: ContentChild,
                args: [NgtGeometry, { static: true }]
            }] } });
class NgtWithGeometryControllerModule {
}
NgtWithGeometryControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtWithGeometryControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryControllerModule, declarations: [NgtWithGeometryController], exports: [NgtWithGeometryController] });
NgtWithGeometryControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryControllerModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithGeometryControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtWithGeometryController],
                    exports: [NgtWithGeometryController],
                }]
        }] });
const [NGT_WITH_GEOMETRY_WATCHED_CONTROLLER, NGT_WITH_GEOMETRY_CONTROLLER_PROVIDER,] = createControllerProviderFactory({
    controller: NgtWithGeometryController,
    watchedControllerTokenName: 'Watched WithGeometryController',
});

class NgtMaterial extends NgtStore {
    constructor(zone, canvasStore, parentObjectFactory) {
        super();
        this.zone = zone;
        this.canvasStore = canvasStore;
        this.parentObjectFactory = parentObjectFactory;
        this.ready = this.select((s) => s.material);
        this.init = this.effect(tapEffect(() => {
            var _a;
            const material = new this.materialType();
            const parentObject = (_a = this.parentObjectFactory) === null || _a === void 0 ? void 0 : _a.call(this);
            if (parentObject) {
                if (Array.isArray(parentObject.material)) {
                    parentObject.material.push(material);
                }
                else {
                    parentObject.material = material;
                }
            }
            this.set({ material });
            return () => {
                if (material) {
                    material.dispose();
                }
            };
        }));
        this.setParameters = this.effect(tap(({ materialParameters, material }) => {
            this.convertColorToLinear(materialParameters);
            material.setValues(Object.assign(materialParameters, 'uniforms' in material && 'uniforms' in materialParameters
                ? {
                    uniforms: Object.assign(Object.assign({}, material.uniforms), materialParameters.uniforms),
                }
                : {}));
            material.needsUpdate = true;
        }));
    }
    set parameters(v) {
        this.set({ materialParameters: v });
    }
    get parameters() {
        return this.get((s) => s.materialParameters);
    }
    get material() {
        return this.get((s) => s.material);
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            this.onCanvasReady(this.canvasStore.ready$, () => {
                this.init();
                this.setParameters(this.select(this.select((s) => s.material), this.select((s) => s.materialParameters), (material, materialParameters) => ({
                    material,
                    materialParameters,
                })));
            });
        });
    }
    convertColorToLinear(parameters) {
        if ('color' in parameters) {
            const colorParams = parameters['color'];
            parameters['color'] = makeColor(colorParams);
            if (!this.canvasStore.isLinear) {
                parameters['color'].convertSRGBToLinear();
            }
        }
    }
}
NgtMaterial.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterial, deps: [{ token: i0.NgZone }, { token: NgtCanvasStore }, { token: NGT_PARENT_OBJECT, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtMaterial.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtMaterial, inputs: { parameters: "parameters" }, outputs: { ready: "ready" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterial, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtCanvasStore }, { type: undefined, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_PARENT_OBJECT]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }], parameters: [{
                type: Input
            }] } });

// GENERATED
class NgtWithMaterialController extends Controller {
    constructor(zone) {
        super();
        this.zone = zone;
        this._material = undefined;
    }
    set material(v) {
        if (v) {
            if (!(Array.isArray(v) && !v.length)) {
                this._materialInput = v;
            }
            this.construct();
        }
    }
    get material() {
        return this._material;
    }
    set contentMaterial(dir) {
        if (dir) {
            this.zone.runOutsideAngular(() => {
                requestAnimationFrame(() => {
                    this.material = dir.material;
                });
            });
        }
    }
    construct() {
        this._material = NgtWithMaterialController.getMaterial(this._materialInput);
    }
    static getMaterial(input) {
        if (Array.isArray(input)) {
            if (!input.length)
                return undefined;
            if (input[0] instanceof THREE.Material) {
                return input;
            }
        }
        if (input instanceof THREE.Material) {
            return input;
        }
        return undefined;
    }
}
NgtWithMaterialController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialController, deps: [{ token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtWithMaterialController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtWithMaterialController, selector: "\n    ngt-mesh,\n    ngt-instanced-mesh,\n    ngt-skinned-mesh,\n    ngt-line,\n    ngt-line-loop,\n    ngt-line-segments,\n    ngt-soba-text,,\n    ngt-points\n  ", inputs: { material: "material" }, queries: [{ propertyName: "contentMaterial", first: true, predicate: NgtMaterial, descendants: true, static: true }], exportAs: ["ngtWithMaterialController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-mesh,
    ngt-instanced-mesh,
    ngt-skinned-mesh,
    ngt-line,
    ngt-line-loop,
    ngt-line-segments,
    ngt-soba-text,,
    ngt-points
  `,
                    exportAs: 'ngtWithMaterialController',
                }]
        }], ctorParameters: function () { return [{ type: i0.NgZone }]; }, propDecorators: { material: [{
                type: Input
            }], contentMaterial: [{
                type: ContentChild,
                args: [NgtMaterial, { static: true }]
            }] } });
class NgtWithMaterialControllerModule {
}
NgtWithMaterialControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtWithMaterialControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialControllerModule, declarations: [NgtWithMaterialController], exports: [NgtWithMaterialController] });
NgtWithMaterialControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialControllerModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtWithMaterialControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtWithMaterialController],
                    exports: [NgtWithMaterialController],
                }]
        }] });
const [NGT_WITH_MATERIAL_WATCHED_CONTROLLER, NGT_WITH_MATERIAL_CONTROLLER_PROVIDER,] = createControllerProviderFactory({
    controller: NgtWithMaterialController,
    watchedControllerTokenName: 'Watched WithMaterialController',
});

// GENERATED
class NgtMaterialGeometryController extends Controller {
    constructor(zone, objectController, objectType, objectPostInit, withMaterialController, withGeometryController) {
        super();
        this.objectController = objectController;
        this._meshArgs = [];
        this.isMultipleMaterials = false;
        objectController.initFn = () => {
            if (!withMaterialController.material) {
                withMaterialController.construct();
            }
            let object;
            if (this.isMultipleMaterials) {
                object = new objectType(withGeometryController.geometry, Array.isArray(withMaterialController.material)
                    ? withMaterialController.material
                    : withMaterialController.material
                        ? [withMaterialController.material]
                        : [], ...this.meshArgs);
            }
            else {
                object = new objectType(withGeometryController.geometry, withMaterialController.material, ...this.meshArgs);
            }
            if (this.morphTargetDictionary &&
                'morphTargetDictionary' in object) {
                object['morphTargetDictionary'] =
                    this.morphTargetDictionary;
            }
            if (this.morphTargetInfluences &&
                'morphTargetInfluences' in object) {
                object['morphTargetInfluences'] =
                    this.morphTargetInfluences;
            }
            if (objectPostInit) {
                zone.runOutsideAngular(() => {
                    objectPostInit(object);
                });
            }
            return object;
        };
        objectController.readyFn = () => {
            if (this.readyFn) {
                this.readyFn();
            }
        };
    }
    set meshArgs(v) {
        this._meshArgs = Array.isArray(v) ? v : [v];
    }
    get meshArgs() {
        return this._meshArgs;
    }
    ngOnInit() {
        this.objectController.init();
    }
    get object() {
        return this.objectController.object;
    }
}
NgtMaterialGeometryController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryController, deps: [{ token: i0.NgZone }, { token: NGT_OBJECT_WATCHED_CONTROLLER }, { token: NGT_OBJECT_TYPE }, { token: NGT_OBJECT_POST_INIT }, { token: NGT_WITH_MATERIAL_WATCHED_CONTROLLER }, { token: NGT_WITH_GEOMETRY_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtMaterialGeometryController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtMaterialGeometryController, selector: "\n    ngt-mesh,\n    ngt-instanced-mesh,\n    ngt-skinned-mesh,\n    ngt-line,\n    ngt-line-loop,\n    ngt-line-segments,\n    ngt-points\n  ", inputs: { isMultipleMaterials: "isMultipleMaterials", morphTargetInfluences: "morphTargetInfluences", morphTargetDictionary: "morphTargetDictionary" }, providers: [
        NGT_OBJECT_CONTROLLER_PROVIDER,
        NGT_WITH_MATERIAL_CONTROLLER_PROVIDER,
        NGT_WITH_GEOMETRY_CONTROLLER_PROVIDER,
    ], exportAs: ["ngtMaterialGeometryController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-mesh,
    ngt-instanced-mesh,
    ngt-skinned-mesh,
    ngt-line,
    ngt-line-loop,
    ngt-line-segments,
    ngt-points
  `,
                    exportAs: 'ngtMaterialGeometryController',
                    providers: [
                        NGT_OBJECT_CONTROLLER_PROVIDER,
                        NGT_WITH_MATERIAL_CONTROLLER_PROVIDER,
                        NGT_WITH_GEOMETRY_CONTROLLER_PROVIDER,
                    ],
                }]
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_TYPE]
                    }] }, { type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_POST_INIT]
                    }] }, { type: NgtWithMaterialController, decorators: [{
                        type: Inject,
                        args: [NGT_WITH_MATERIAL_WATCHED_CONTROLLER]
                    }] }, { type: NgtWithGeometryController, decorators: [{
                        type: Inject,
                        args: [NGT_WITH_GEOMETRY_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { isMultipleMaterials: [{
                type: Input
            }], morphTargetInfluences: [{
                type: Input
            }], morphTargetDictionary: [{
                type: Input
            }] } });
class NgtMaterialGeometryControllerModule {
}
NgtMaterialGeometryControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtMaterialGeometryControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryControllerModule, declarations: [NgtMaterialGeometryController], exports: [NgtMaterialGeometryController, NgtObjectControllerModule,
        NgtWithMaterialControllerModule,
        NgtWithGeometryControllerModule] });
NgtMaterialGeometryControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryControllerModule, imports: [NgtObjectControllerModule,
        NgtWithMaterialControllerModule,
        NgtWithGeometryControllerModule] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtMaterialGeometryControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtMaterialGeometryController],
                    exports: [
                        NgtMaterialGeometryController,
                        NgtObjectControllerModule,
                        NgtWithMaterialControllerModule,
                        NgtWithGeometryControllerModule,
                    ],
                }]
        }] });
const [NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER, NGT_MATERIAL_GEOMETRY_CONTROLLER_PROVIDER,] = createControllerProviderFactory({
    watchedControllerTokenName: 'Watched MaterialGeometryController',
    controller: NgtMaterialGeometryController,
});

// GENERATED
class NgtAudioController extends Controller {
    constructor() {
        super(...arguments);
        this.buffer = null;
        this.detune = 0;
        this.loop = false;
        this.loopStart = 0;
        this.loopEnd = 0;
        this.offset = 0;
        this.duration = undefined;
        this.playbackRate = 1;
        this.isPlaying = false;
        this.hasPlaybackControl = true;
        this.sourceType = 'empty';
        this.source = null;
        this.filters = [];
    }
}
NgtAudioController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioController, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtAudioController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtAudioController, selector: "\n    ngt-audio,\n    ngt-positional-audio,\n    ngt-soba-positional-audio,\n  ", inputs: { autoplay: "autoplay", buffer: "buffer", detune: "detune", loop: "loop", loopStart: "loopStart", loopEnd: "loopEnd", offset: "offset", duration: "duration", playbackRate: "playbackRate", isPlaying: "isPlaying", hasPlaybackControl: "hasPlaybackControl", sourceType: "sourceType", source: "source", filters: "filters" }, exportAs: ["ngtAudioController"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-audio,
    ngt-positional-audio,
    ngt-soba-positional-audio,
  `,
                    exportAs: 'ngtAudioController',
                }]
        }], propDecorators: { autoplay: [{
                type: Input
            }], buffer: [{
                type: Input
            }], detune: [{
                type: Input
            }], loop: [{
                type: Input
            }], loopStart: [{
                type: Input
            }], loopEnd: [{
                type: Input
            }], offset: [{
                type: Input
            }], duration: [{
                type: Input
            }], playbackRate: [{
                type: Input
            }], isPlaying: [{
                type: Input
            }], hasPlaybackControl: [{
                type: Input
            }], sourceType: [{
                type: Input
            }], source: [{
                type: Input
            }], filters: [{
                type: Input
            }] } });
class NgtAudioControllerModule {
}
NgtAudioControllerModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioControllerModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtAudioControllerModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioControllerModule, declarations: [NgtAudioController], exports: [NgtAudioController] });
NgtAudioControllerModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioControllerModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAudioControllerModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtAudioController],
                    exports: [NgtAudioController],
                }]
        }] });
const [NGT_AUDIO_WATCHED_CONTROLLER, NGT_AUDIO_CONTROLLER_PROVIDER] = createControllerProviderFactory({
    watchedControllerTokenName: 'Watched AudioController',
    controller: NgtAudioController,
});

class NgtCommonAudio {
    constructor(objectController) {
        this.objectController = objectController;
        this.ready = new EventEmitter();
        this._audioArgs = [];
        objectController.initFn = () => {
            if (!this.listener) {
                throw new Error('Cannot initialize Audio without an AudioListener');
            }
            return (this._audio = new this.audioType(this.listener, ...this._audioArgs));
        };
        objectController.readyFn = () => {
            this.ready.emit(this.audio);
        };
    }
    set audioArgs(v) {
        this._audioArgs = Array.isArray(v) ? v : [v];
        this.objectController.init();
    }
    ngOnInit() {
        if (!this._audio) {
            this.objectController.init();
        }
    }
    get audio() {
        return this._audio;
    }
}
NgtCommonAudio.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonAudio, deps: [{ token: NGT_OBJECT_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonAudio.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCommonAudio, inputs: { listener: "listener" }, outputs: { ready: "ready" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonAudio, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { listener: [{
                type: Input
            }], ready: [{
                type: Output
            }] } });

class NgtCommonLine {
    constructor(materialGeometryController) {
        this.materialGeometryController = materialGeometryController;
        this.ready = new EventEmitter();
        if (materialGeometryController) {
            materialGeometryController.readyFn = () => {
                this.ready.emit(this.line);
            };
        }
    }
    get line() {
        var _a;
        return (_a = this.materialGeometryController) === null || _a === void 0 ? void 0 : _a.object;
    }
}
NgtCommonLine.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonLine, deps: [{ token: NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonLine.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCommonLine, outputs: { ready: "ready" }, providers: [{ provide: NGT_OBJECT_TYPE, useValue: THREE.Line }], ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonLine, decorators: [{
            type: Directive,
            args: [{
                    providers: [{ provide: NGT_OBJECT_TYPE, useValue: THREE.Line }],
                }]
        }], ctorParameters: function () {
        return [{ type: NgtMaterialGeometryController, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtTexture extends NgtStore {
    constructor(zone, canvasStore) {
        super();
        this.zone = zone;
        this.canvasStore = canvasStore;
        this.ready = new EventEmitter();
        this._textureArgs = [];
    }
    set textureArgs(v) {
        this._textureArgs = Array.isArray(v) ? v : [v];
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            this.initSubscription = this.onCanvasReady(this.canvasStore.ready$, () => {
                this._texture = new this.textureType(...this._textureArgs);
                this.set({ texture: this.texture });
            });
        });
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            this.effect(tap((texture) => {
                this.ready.emit(texture);
            }))(this.select((s) => s.texture));
        });
    }
    get texture() {
        return this._texture;
    }
    ngOnDestroy() {
        this.zone.runOutsideAngular(() => {
            if (this.texture) {
                this.texture.dispose();
            }
        });
        super.ngOnDestroy();
    }
}
NgtTexture.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtTexture, deps: [{ token: i0.NgZone }, { token: NgtCanvasStore }], target: i0.ɵɵFactoryTarget.Directive });
NgtTexture.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtTexture, outputs: { ready: "ready" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtTexture, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: i0.NgZone }, { type: NgtCanvasStore }]; }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtLight {
    constructor(objectController) {
        this.objectController = objectController;
        this.ready = new EventEmitter();
        this.initializing = false;
        this._lightArgs = [];
        objectController.initFn = () => {
            this._light = new this.lightType(...this._lightArgs);
            if (this.intensity) {
                applyProps(this.light, { intensity: this.intensity });
            }
            if (this.shadow) {
                applyProps(this.light, this.shadow);
            }
            this.initializing = false;
            return this.light;
        };
        objectController.readyFn = () => {
            this.ready.emit(this.light);
        };
    }
    set lightArgs(v) {
        this.initializing = true;
        this._lightArgs = Array.isArray(v) ? v : [v];
        this.objectController.init();
    }
    ngOnInit() {
        if (!this._light && !this.initializing) {
            this.objectController.init();
        }
    }
    get light() {
        return this._light;
    }
}
NgtLight.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLight, deps: [{ token: NGT_OBJECT_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtLight.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtLight, inputs: { intensity: "intensity", shadow: "shadow" }, outputs: { ready: "ready" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLight, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }], intensity: [{
                type: Input
            }], shadow: [{
                type: Input
            }] } });

class NgtAttribute extends NgtStore {
    constructor(zone, geometryDirective, canvasStore) {
        super();
        this.zone = zone;
        this.geometryDirective = geometryDirective;
        this.canvasStore = canvasStore;
        this._attributeArgs = [];
    }
    set attributeArgs(v) {
        this._attributeArgs = Array.isArray(v) ? v : [v];
        this.init();
    }
    ngOnChanges() {
        this.zone.runOutsideAngular(() => {
            if (this.attribute) {
                this.attribute.needsUpdate = true;
            }
        });
    }
    ngOnInit() {
        if (!this.attribute) {
            this.init();
        }
    }
    init() {
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            this.initSubscription = this.onCanvasReady(this.canvasStore.ready$, () => {
                if (this.geometryDirective && this.attach) {
                    this._attribute = new this.attributeType(...this._attributeArgs);
                    if (this.attribute) {
                        this._defaultValue = this.geometryDirective.geometry
                            .attributes[this.attach];
                        this.geometryDirective.geometry.setAttribute(this.attach, this.attribute);
                    }
                }
            });
        });
    }
    ngOnDestroy() {
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            if (this.geometryDirective && this.attach) {
                if (this._defaultValue !== undefined) {
                    this.geometryDirective.geometry.setAttribute(this.attach, this._defaultValue);
                }
                else {
                    this.geometryDirective.geometry.deleteAttribute(this.attach);
                }
            }
        });
        super.ngOnDestroy();
    }
    get attribute() {
        return this._attribute;
    }
}
NgtAttribute.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAttribute, deps: [{ token: i0.NgZone }, { token: NgtGeometry, optional: true }, { token: NgtCanvasStore }], target: i0.ɵɵFactoryTarget.Directive });
NgtAttribute.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtAttribute, inputs: { attach: "attach" }, usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtAttribute, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtGeometry, decorators: [{
                        type: Optional
                    }] }, { type: NgtCanvasStore }];
    }, propDecorators: { attach: [{
                type: Input
            }] } });

class NgtCurve extends NgtStore {
    constructor(zone, geometryDirective, canvasStore) {
        super();
        this.zone = zone;
        this.geometryDirective = geometryDirective;
        this.canvasStore = canvasStore;
        this._curveArgs = [];
    }
    set curveArgs(v) {
        this._curveArgs = Array.isArray(v) ? v : [v];
        this.init();
    }
    ngOnInit() {
        if (!this.curve) {
            this.init();
        }
    }
    init() {
        this.zone.runOutsideAngular(() => {
            if (this.initSubscription) {
                this.initSubscription.unsubscribe();
            }
            this.initSubscription = this.onCanvasReady(this.canvasStore.ready$, () => {
                this._curve = new this.curveType(...this._curveArgs);
                if (this.curve && this.geometryDirective) {
                    const points = this.curve.getPoints(this.divisions);
                    this.geometryDirective.geometry.setFromPoints(points);
                }
            });
        });
    }
    get curve() {
        return this._curve;
    }
}
NgtCurve.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCurve, deps: [{ token: i0.NgZone }, { token: NgtGeometry, optional: true }, { token: NgtCanvasStore }], target: i0.ɵɵFactoryTarget.Directive });
NgtCurve.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCurve, inputs: { divisions: "divisions" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCurve, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: i0.NgZone }, { type: NgtGeometry, decorators: [{
                        type: Optional
                    }] }, { type: NgtCanvasStore }];
    }, propDecorators: { divisions: [{
                type: Input
            }] } });

class NgtHelper {
    constructor(objectController) {
        this.objectController = objectController;
        this.ready = new EventEmitter();
        this._helperArgs = [];
        objectController.initFn = () => {
            return (this._helper = new this.helperType(...this._helperArgs));
        };
        objectController.readyFn = () => {
            this.ready.emit(this.helper);
        };
    }
    set helperArgs(v) {
        this._helperArgs = Array.isArray(v) ? v : [v];
        this.objectController.init();
    }
    ngOnChanges() {
        this.objectController.init();
    }
    ngOnInit() {
        if (!this._helper) {
            this.objectController.init();
        }
    }
    get helper() {
        return this._helper;
    }
}
NgtHelper.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtHelper, deps: [{ token: NGT_OBJECT_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtHelper.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtHelper, outputs: { ready: "ready" }, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtHelper, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtCommonMesh {
    constructor(materialGeometryController) {
        this.materialGeometryController = materialGeometryController;
        this.ready = new EventEmitter();
        if (materialGeometryController) {
            materialGeometryController.readyFn = () => {
                this.ready.emit(this.mesh);
            };
        }
    }
    get mesh() {
        var _a;
        return (_a = this.materialGeometryController) === null || _a === void 0 ? void 0 : _a.object;
    }
}
NgtCommonMesh.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonMesh, deps: [{ token: NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonMesh.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCommonMesh, outputs: { ready: "ready" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonMesh, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtMaterialGeometryController, decorators: [{
                        type: Optional
                    }, {
                        type: Inject,
                        args: [NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtCommonSprite {
    constructor(objectController) {
        this.objectController = objectController;
        this.ready = new EventEmitter();
        objectController.initFn = () => {
            if (this.material) {
                this._sprite = new this.spriteType(this.material);
            }
            else if (this.ngtMaterial) {
                if (this.ngtMaterial.material instanceof THREE.SpriteMaterial) {
                    this._sprite = new this.spriteType(this.ngtMaterial.material);
                }
            }
            return this._sprite;
        };
        objectController.readyFn = () => {
            this.ready.emit(this.sprite);
        };
    }
    ngOnInit() {
        this.objectController.init();
    }
    get sprite() {
        return this._sprite;
    }
}
NgtCommonSprite.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonSprite, deps: [{ token: NGT_OBJECT_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonSprite.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCommonSprite, inputs: { material: "material" }, outputs: { ready: "ready" }, queries: [{ propertyName: "ngtMaterial", first: true, predicate: NgtMaterial, descendants: true }], ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonSprite, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { material: [{
                type: Input
            }], ready: [{
                type: Output
            }], ngtMaterial: [{
                type: ContentChild,
                args: [NgtMaterial]
            }] } });

class NgtCommonCamera {
    constructor(objectController) {
        this.objectController = objectController;
        this.ready = new EventEmitter();
        this._cameraArgs = [];
        objectController.initFn = () => {
            return (this._camera = new this.cameraType(...this._cameraArgs));
        };
        objectController.readyFn = () => {
            this.ready.emit(this.camera);
        };
    }
    set cameraArgs(v) {
        this._cameraArgs = Array.isArray(v) ? v : [v];
        this.objectController.init();
    }
    ngOnInit() {
        if (!this._camera) {
            this.objectController.init();
        }
    }
    get camera() {
        return this._camera;
    }
}
NgtCommonCamera.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonCamera, deps: [{ token: NGT_OBJECT_WATCHED_CONTROLLER }], target: i0.ɵɵFactoryTarget.Directive });
NgtCommonCamera.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtCommonCamera, outputs: { ready: "ready" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtCommonCamera, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: NgtObjectController, decorators: [{
                        type: Inject,
                        args: [NGT_OBJECT_WATCHED_CONTROLLER]
                    }] }];
    }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtExtender {
    constructor() {
        this.ready = new EventEmitter();
        this.animateReady = new EventEmitter();
    }
    set object(value) {
        this._object = value;
        this.ready.emit(value);
    }
    get object() {
        return this._object;
    }
}
NgtExtender.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtExtender, deps: [], target: i0.ɵɵFactoryTarget.Directive });
NgtExtender.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtExtender, outputs: { ready: "ready", animateReady: "animateReady" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtExtender, decorators: [{
            type: Directive
        }], propDecorators: { ready: [{
                type: Output
            }], animateReady: [{
                type: Output
            }] } });
function createExtenderProvider(type) {
    return {
        provide: NgtExtender,
        useExisting: type,
    };
}

class NgtObjectHelper extends NgtStore {
    constructor(objectFn, canvasStore, animationFrameStore, zone) {
        super();
        this.objectFn = objectFn;
        this.canvasStore = canvasStore;
        this.animationFrameStore = animationFrameStore;
        this.zone = zone;
    }
    set objectHelperArgs(v) {
        this.set({ args: Array.isArray(v) ? v : [v] });
    }
    ngOnInit() {
        this.zone.runOutsideAngular(() => {
            this.effect(tapEffect((args) => {
                this._object = this.objectFn();
                if (!this._object) {
                    console.info('Parent is not an object3d');
                    return;
                }
                this._objectHelper = new this.objectHelperType(this._object, ...args);
                const scene = this.canvasStore.get((s) => s.scene);
                if (this.objectHelper && scene) {
                    scene.add(this.objectHelper);
                    const animationUuid = this.animationFrameStore.register({
                        callback: () => {
                            if (this.objectHelper) {
                                this
                                    .objectHelper.update();
                            }
                        },
                    });
                    return () => {
                        if (this.objectHelper && scene) {
                            scene.remove(this.objectHelper);
                            this.animationFrameStore.unregister(animationUuid);
                        }
                    };
                }
                return;
            }))(this.select(this.select((s) => s.args), this.canvasStore.ready$, (args) => args));
        });
    }
    get objectHelper() {
        return this._objectHelper;
    }
}
NgtObjectHelper.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectHelper, deps: [{ token: NGT_PARENT_OBJECT }, { token: NgtCanvasStore }, { token: NgtAnimationFrameStore }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtObjectHelper.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "13.2.2", type: NgtObjectHelper, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtObjectHelper, decorators: [{
            type: Directive
        }], ctorParameters: function () {
        return [{ type: undefined, decorators: [{
                        type: Inject,
                        args: [NGT_PARENT_OBJECT]
                    }] }, { type: NgtCanvasStore }, { type: NgtAnimationFrameStore }, { type: i0.NgZone }];
    } });

function buildGraph(object) {
    const data = { nodes: {}, materials: {} };
    if (object) {
        object.traverse((obj) => {
            if (obj.name) {
                data.nodes[obj.name] = obj;
            }
            if (obj['material'] &&
                !data.materials[obj['material']['name']]) {
                data.materials[obj['material']['name']] = obj['material'];
            }
        });
    }
    return data;
}

class NgtLoader {
    constructor() {
        this.cached = new Map();
    }
    use(loaderConstructor, input, extensions, onProgress) {
        const keys = (Array.isArray(input) ? input : [input]);
        const loader = new loaderConstructor();
        if (extensions) {
            extensions(loader);
        }
        const results$ = forkJoin(keys.map((key) => {
            if (this.cached.has(key)) {
                return of(this.cached.get(key));
            }
            return defer(() => loader.loadAsync(key, onProgress)).pipe(tap((data) => {
                if (data.scene) {
                    Object.assign(data, buildGraph(data.scene));
                }
                this.cached.set(key, data);
            }), catchError((err) => {
                console.error(`Error loading ${key}: ${err.message}`);
                return throwError(err);
            }));
        }));
        return defer(() => Array.isArray(input)
            ? results$
            : results$.pipe(map((results) => results[0]))).pipe(share({
            connector: () => new ReplaySubject(),
            resetOnRefCountZero: true,
            resetOnError: true,
        }));
    }
    ngOnDestroy() {
        this.cached.clear();
    }
}
NgtLoader.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoader, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgtLoader.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoader, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "13.2.2", ngImport: i0, type: NgtLoader, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }] });

function shaderMaterial(uniforms, vertexShader, fragmentShader, onInit) {
    return class extends THREE.ShaderMaterial {
        constructor() {
            const entries = Object.entries(uniforms);
            // Create unforms and shaders
            super({
                uniforms: entries.reduce((acc, [name, value]) => {
                    const uniform = THREE.UniformsUtils.clone({
                        [name]: { value },
                    });
                    return Object.assign(Object.assign({}, acc), uniform);
                }, {}),
                vertexShader,
                fragmentShader,
            });
            // Create getter/setters
            entries.forEach(([name]) => Object.defineProperty(this, name, {
                get: () => this.uniforms[name].value,
                set: (v) => (this.uniforms[name].value = v),
            }));
            if (onInit)
                onInit(this);
        }
    };
}

/**
 * Generated bundle index. Do not edit.
 */

export { Controller, NGT_AUDIO_CONTROLLER_PROVIDER, NGT_AUDIO_WATCHED_CONTROLLER, NGT_CANVAS_OPTIONS, NGT_HOST_PARENT_OBJECT, NGT_IS_WEBGL_AVAILABLE, NGT_MATERIAL_GEOMETRY_CONTROLLER_PROVIDER, NGT_MATERIAL_GEOMETRY_WATCHED_CONTROLLER, NGT_OBJECT_CONTROLLER_PROVIDER, NGT_OBJECT_INPUTS_CONTROLLER_PROVIDER, NGT_OBJECT_INPUTS_WATCHED_CONTROLLER, NGT_OBJECT_POST_INIT, NGT_OBJECT_TYPE, NGT_OBJECT_WATCHED_CONTROLLER, NGT_PARENT_OBJECT, NGT_PERFORMANCE_OPTIONS, NGT_RESIZE_BOX_DEFAULT, NGT_RESIZE_DEBOUNCE_DEFAULT, NGT_RESIZE_OBSERVER_SUPPORT, NGT_RESIZE_OPTIONS, NGT_WITH_GEOMETRY_CONTROLLER_PROVIDER, NGT_WITH_GEOMETRY_WATCHED_CONTROLLER, NGT_WITH_MATERIAL_CONTROLLER_PROVIDER, NGT_WITH_MATERIAL_WATCHED_CONTROLLER, NgtAnimationFrameStore, NgtAttribute, NgtAudioController, NgtAudioControllerModule, NgtCanvas, NgtCanvasStore, NgtColorPipe, NgtColorPipeModule, NgtCommonAudio, NgtCommonCamera, NgtCommonLine, NgtCommonMesh, NgtCommonSprite, NgtCoreModule, NgtCursor, NgtCursorModule, NgtCurve, NgtDestroyed, NgtEventsStore, NgtExtender, NgtFogExp2Pipe, NgtFogPipe, NgtFogPipeModule, NgtGeometry, NgtHelper, NgtLight, NgtLoader, NgtLoop, NgtMaterial, NgtMaterialGeometryController, NgtMaterialGeometryControllerModule, NgtMathConstantPipe, NgtMathPipe, NgtMathPipeModule, NgtObjectController, NgtObjectControllerModule, NgtObjectHelper, NgtObjectInputsController, NgtObjectInputsControllerModule, NgtPerformance, NgtPiPipe, NgtPiPipeModule, NgtRadianPipe, NgtRadianPipeModule, NgtRepeat, NgtRepeatModule, NgtResize, NgtSidePipe, NgtSidePipeModule, NgtStore, NgtTexture, NgtVector2Pipe, NgtVector3Pipe, NgtVector4Pipe, NgtVectorPipeModule, NgtWithGeometryController, NgtWithGeometryControllerModule, NgtWithMaterialController, NgtWithMaterialControllerModule, applyProps, buildGraph, controllerFactory, createControllerProviderFactory, createExtenderProvider, createHostParentObjectProvider, createParentObjectProvider, debounceSync, isOrthographicCamera, makeColor, makeForSet, makeId, makeVector2, makeVector3, makeVector4, provideCanvasOptions, shaderMaterial, skipUndefined, startWithUndefined, tapEffect };
//# sourceMappingURL=angular-three-core.mjs.map
